#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing the core DROP classes.
"""
import ast
import inspect
import logging
import os
import threading
import time
import re
import sys
from abc import ABCMeta

from dlg.common.reproducibility.constants import (
    ReproducibilityFlags,
    REPRO_DEFAULT,
    rmode_supported,
    ALL_RMODES,
)
from dlg.common.reproducibility.reproducibility import common_hash
from merklelib import MerkleTree

from .ddap_protocol import (
    ExecutionMode,
    ChecksumTypes,
    DROPLinkType,
    DROPPhases,
    DROPStates,
    DROPRel,
)
from dlg.event import EventFirer, EventHandler
from dlg.exceptions import InvalidDropException, InvalidRelationshipException

DEFAULT_INTERNAL_PARAMETERS = {
    "dropclass",
    "category",
    "storage",
    "fields",
    "streaming",
    "persist",
    "rank",
    "loop_ctx",
    "weight",
    "iid",
    "consumers",
    "config_data",
    "mode",
    "group_end",
    "applicationArgs",
    "reprodata",
}

if sys.version_info >= (3, 8):
    pass
from dlg.utils import (
    object_tracking,
    getDlgVariable,
    truncateUidToKey,
)
from dlg.meta import (
    dlg_float_param,
    dlg_int_param,
    dlg_list_param,
    dlg_string_param,
    dlg_enum_param,
    dlg_bool_param,
    dlg_dict_param,
)

# Opt into using per-drop checksum calculation
checksum_disabled = "DLG_DISABLE_CHECKSUM" in os.environ
try:
    from crc32c import crc32c  # pylint: disable=unused-import

    _checksumType = ChecksumTypes.CRC_32C
except (ModuleNotFoundError, ImportError):
    from binascii import crc32 # pylint: disable=unused-import

    _checksumType = ChecksumTypes.CRC_32

logger = logging.getLogger(f"dlg.{__name__}")


class ListAsDict(list):
    """A list that adds drop UIDs to a set as they get appended to the list"""

    def __init__(self, my_set):
        self.set = my_set

    def append(self, drop):
        super(ListAsDict, self).append(drop)
        self.set.add(drop.uid)


track_current_drop = object_tracking("drop")


# ===============================================================================
# DROP classes follow
# ===============================================================================


class AbstractDROP(EventFirer, EventHandler):
    """
    Base class for all DROP implementations.

    A DROP is a representation of a piece of data. DROPs are created,
    written once, potentially read many times, and they finally potentially
    expire and get deleted. Subclasses implement different storage mechanisms
    to hold the data represented by the DROP.

    If the data represented by this DROP is written *through* this object
    (i.e., calling the ``write`` method), this DROP will keep track of the
    data's size and checksum. If the data is written externally, the size and
    checksum can be fed into this object for future reference.

    DROPs can have consumers attached to them. 'Normal' consumers will
    wait until the DROP they 'consume' (their 'input') moves to the
    COMPLETED state and then will consume it, most typically by opening it
    and reading its contents, but any other operation could also be performed.
    How the consumption is triggered depends on the producer's `executionMode`
    flag, which dictates whether it should trigger the consumption itself or
    if it should be manually triggered by an external entity. On the other hand,
    streaming consumers receive the data that is written into its input
    *as it gets written*. This mechanism is driven always by the DROP that
    acts as a streaming input. Apart from receiving the data as it gets
    written into the DROP, streaming consumers are also notified when the
    DROPs moves to the COMPLETED state, at which point no more data should
    be expected to arrive at the consumer side.

    DROPs' data can be expired automatically by the system after the DROP has
    transitioned to the COMPLETED state if they are created by a DROP Manager.
    Expiration can either be triggered by an interval relative to the creation
    time of the DROP (via the `lifespan` keyword), or by specifying that the
    DROP should be expired after all its consumers have finished (via the
    `expireAfterUse` keyword). These two methods are mutually exclusive. If none
    is specified no expiration occurs.
    """

    # This ensures that:
    #  - This class cannot be instantiated
    #  - Subclasses implement methods decorated with @abstractmethod
    __metaclass__ = ABCMeta

    # Matcher used to validate environment_variable_syntax
    _env_var_matcher = re.compile(r"\$[A-z|\d]+\..+")
    _dlg_var_matcher = re.compile(r"\$DLG_.+")

    _known_locks = ("_finishedProducersLock", "_refLock")
    _known_rlocks = ("_statusLock",)

    def __getstate__(self):
        state = self.__dict__.copy()
        for attr_name in AbstractDROP._known_locks + AbstractDROP._known_rlocks:
            del state[attr_name]
        del state["_listeners"]
        return state

    def __setstate__(self, state):
        for attr_name in AbstractDROP._known_locks:
            state[attr_name] = threading.Lock()
        for attr_name in AbstractDROP._known_rlocks:
            state[attr_name] = threading.RLock()

        self.__dict__.update(state)

        import collections

        self._listeners = collections.defaultdict(list)

    @track_current_drop
    def __init__(self, oid, uid, **kwargs):
        """
        Creates a DROP. The only mandatory argument are the Object ID
        (`oid`) and the Unique ID (`uid`) of the new object (see the `self.oid`
        and `self.uid` methods for more information). Any extra arguments must
        be keyed, and will be processed either by this method, or by the
        `initialize` method.

        This method should not be overwritten by subclasses. For any specific
        initialization logic, the `initialize` method should be overwritten
        instead. This allows us to move to the INITIALIZED state only after any
        specific initialization has occurred in the subclasses.
        """

        super(AbstractDROP, self).__init__()

        if "_dlg_session_id" not in kwargs and "dlg_session_id" in kwargs:
            kwargs["_dlg_session_id"] = kwargs["dlg_session_id"]
        self._extract_attributes(**kwargs)

        # Copy it since we're going to modify it
        kwargs = dict(kwargs)
        # So far only these three are mandatory
        self._oid = str(oid)
        self._uid = str(uid)

        # Set log_level for this drop to level provided
        self._log_level = self._popArg(
            kwargs, "log_level", logging.getLevelName(logger.getEffectiveLevel())
        )
        if self._log_level == "" or self._log_level == logging.getLevelName(
                logging.NOTSET
        ):
            self._log_level = logging.getLevelName(logger.getEffectiveLevel())
        self._global_log_level = logging.getLevelName(logger.getEffectiveLevel())

        # The physical graph drop type. This is determined
        # by the drop category when generating the drop spec
        self._type = self._popArg(kwargs, "categoryType", None)

        # The ID of the session owning this drop, if any
        # In most real-world situations this attribute will be set, but in
        # general it cannot be assumed it will (e.g., unit tests create drops
        # directly outside the context of a session).
        self._dlg_session_id = self._popArg(kwargs, "dlg_session_id", "")

        # A simple name that the Drop might receive
        # This is usually set in the Logical Graph Editor,
        # but is not necessarily always there
        self.name = self._popArg(kwargs, "name", self._oid)

        # The key of this drop in the original Logical Graph
        # This information might or might not be present depending on how the
        # physical graph was generated (or if this drop is being created as part
        # of a graph, to begin with), so we default it to an empty value
        self.lg_key = self._popArg(kwargs, "lg_key", "")

        # 1-to-N relationship: one DROP may have many consumers and producers.
        # The potential consumers and producers are always AppDROPs instances
        # We keep the UIDs in a set for O(1) "x in set" operations
        # Obviously the normal way of doing this is using a dictionary, but
        # for the time being and while testing the integration with TBU's ceda
        # library we need to expose a list.
        self._consumers_uids = set()
        self._consumers = ListAsDict(self._consumers_uids)
        self._producers_uids = set()
        self._producers = ListAsDict(self._producers_uids)

        # Set holding the state of the producers that have finished their
        # execution. Once all producers have finished, this DROP moves
        # itself to the COMPLETED state
        self._finishedProducers = []
        self._finishedProducersLock = threading.Lock()

        # Streaming consumers are objects that consume the data written in
        # this DROP *as it gets written*, and therefore don't have to
        # wait until this DROP has moved to COMPLETED.
        # An object cannot be a streaming consumers and a 'normal' consumer
        # at the same time, although this rule is imposed simply to enforce
        # efficiency (why would a consumer want to consume the data twice?) and
        # not because it's technically impossible.
        # See comment above in self._consumers/self._producers for separate set
        # with uids
        self._streamingConsumers_uids = set()
        self._streamingConsumers = ListAsDict(self._streamingConsumers_uids)

        self._refCount = 0
        self._refLock = threading.Lock()
        self._location = None
        self._parent = None
        self._status = None
        self._statusLock = threading.RLock()

        # Current and target phases.
        # Phases represent the resiliency of data. An initial phase of PLASMA
        # is set on each DROP representing its lack of non-transient storage
        # support. A target phase is also set to hint the Data Lifecycle Manager
        # about the level of resilience that this DROP should achieve.
        self._phase = DROPPhases.PLASMA
        self._targetPhase = self._popArg(kwargs, "targetPhase", DROPPhases.GAS)

        # Calculating the checksum and maintaining the data size internally
        # implies that the data represented by this DROP is written
        # *through* this DROP. This might not always be the case though,
        # since data could be written externally and the DROP simply be
        # moved to COMPLETED at the end of the process. In this case we return a
        # None checksum and size (when requested), signaling that we don't have
        # this information.
        # Note also that the setters of these two properties also allow to set
        # a value on them, but only if they are None
        self._checksum = None
        self._checksumType = None
        self._size = None

        # Recording runtime reproducibility information is handled via MerkleTrees
        # Switching on the reproducibility level will determine what information is recorded.
        self._committed = False
        self._merkleRoot = None
        self._merkleTree = None
        self._merkleData = []
        self._reproducibility = REPRO_DEFAULT

        # The DataIO instance we use in our write method. It's initialized to
        # None because it's lazily initialized in the write method, since data
        # might be written externally and not through this DROP
        self._wio = None

        # DataIO objects used for reading.
        # Instead of passing file objects or more complex data types in our
        # open/read/close calls we use integers, mainly because Pyro doesn't
        # handle file types and other classes (like StringIO) well, but also
        # because it requires less transport.
        # TODO: Make these threadsafe, no lock around them yet
        self._rios = {}

        self._humanKey = self._popArg(
            kwargs, "humanReadableKey", truncateUidToKey(self._uid)
        )
        # The execution mode.
        # When set to DROP (the default) the graph execution will be driven by
        # DROPs themselves by firing and listening to events, and reacting
        # accordingly by executing themselves or moving to the COMPLETED state.
        # When set to EXTERNAL, DROPs do no react to these events, and remain
        # in the state they currently are. In this case an external entity must
        # listen to the events and decide when to trigger the execution of the
        # applications.
        self._executionMode = self._popArg(kwargs, "executionMode", ExecutionMode.DROP)

        # The physical node where this DROP resides.
        # This piece of information is mandatory when submitting the physical
        # graph via the DataIslandManager, but in simpler scenarios such as
        # tests or graph submissions via the NodeManager it might be
        # missing.
        self._node = self._popArg(kwargs, "node", None)

        # The host representing the Data Island where this DROP resides
        # This piece of information is mandatory when submitting the physical
        # graph via the MasterManager, but in simpler scenarios such as tests or
        # graphs submissions via the DataIslandManager or NodeManager it might
        # missing.
        self._dataIsland = self._popArg(kwargs, "island", None)

        # DROP expiration.
        # Expiration can be time-driven or usage-driven, which are mutually
        # exclusive methods. If time-driven, a relative lifespan is assigned to
        # the DROP.
        # Expected lifespan for this object, used by to expire them
        if "lifespan" in kwargs and "expireAfterUse" in kwargs:
            raise InvalidDropException(
                self,
                "%r specifies both `lifespan` and `expireAfterUse`"
                "but they are mutually exclusive" % (self,),
            )

        self._expireAfterUse = self._popArg(kwargs, "expireAfterUse", False)
        self._expirationDate = -1
        if not self._expireAfterUse:
            lifespan = float(self._popArg(kwargs, "lifespan", -1))
            if lifespan != -1:
                self._expirationDate = time.time() + lifespan

        # Expected data size, used to automatically move the DROP to COMPLETED
        # after successive calls to write()
        self._expectedSize = -1
        if "expectedSize" in kwargs and kwargs["expectedSize"]:
            self._expectedSize = int(kwargs.pop("expectedSize"))

        # No DROP should be persisted unless stated otherwise; used for replication
        self._persist: bool = self._popArg(kwargs, "persist", False)
        # If DROP should be persisted, don't expire (delete) it.
        if self._persist:
            self._expireAfterUse = False

        # Flag to control whether a call to skip blocks until the
        # last producer is finished. This is useful for data drops capturing
        # the output of multiple branches. Default is false, meaning that
        # the first call to skip will close the drop and also skip the following
        # drops.
        self.block_skip = self._popArg(kwargs, "block_skip", False)

        # Useful to have access to all EAGLE parameters without a prior knowledge
        self._parameters = dict(kwargs)
        self.autofill_environment_variables()
        kwargs.update(self._parameters)
        # Sub-class initialization; mark ourselves as INITIALIZED after that
        self.initialize(**kwargs)
        self._status = (
            DROPStates.INITIALIZED
        )  # no need to use synchronised self.status here

    _members_cache = {}

    def _get_members(self):
        cls = self.__class__
        if cls not in AbstractDROP._members_cache:
            members = [
                (name, val)
                for c in cls.__mro__[:-1]
                for name, val in vars(c).items()
                if not (inspect.isfunction(val) or isinstance(val, property))
            ]
            # logger.info("GOT MEMBEEEERS: %r", members)

            AbstractDROP._members_cache[cls] = members
        return AbstractDROP._members_cache[cls]

    def _extract_attributes(self, **kwargs):
        """
        Extracts component and app params then assigns them to class instance attributes.
        Component params take pro
        """

        def get_param_value(attr_name, default_value):

            if attr_name in kwargs:
                return kwargs.get(attr_name)
            elif "applicationArgs" in kwargs and attr_name in kwargs["applicationArgs"] and (
                kwargs["applicationArgs"].get(attr_name).usage not in [
                    "InputPort",
                    "OutputPort",
                    "InputOutput",
                    ]
                ):
                    return kwargs["applicationArgs"].get(attr_name).value
            else:
                return default_value

        # Take a class dlg defined parameter class attribute and create an instanced attribute on object
        for attr_name, member in self._get_members():
            if isinstance(member, dlg_float_param):
                value = get_param_value(attr_name, member.default_value)
                if value is not None and value != "":
                    value = float(value)
            elif isinstance(member, dlg_bool_param):
                value = get_param_value(attr_name, member.default_value)
                if value is not None and value != "":
                    value = bool(value)
            elif isinstance(member, dlg_int_param):
                value = get_param_value(attr_name, member.default_value)
                if value is not None and value != "":
                    value = int(value)
            elif isinstance(member, dlg_string_param):
                value = get_param_value(attr_name, member.default_value)
                if value is not None and value != "":
                    value = str(value)
            elif isinstance(member, dlg_enum_param):
                value = get_param_value(attr_name, member.default_value)
                if value is not None and value != "":
                    value = member.cls(value)
            elif isinstance(member, dlg_list_param):
                value = get_param_value(attr_name, member.default_value)
                if isinstance(value, str):
                    if value == "":
                        value = []
                    else:
                        value = ast.literal_eval(value)
                if value is not None and not isinstance(value, list):
                    raise TypeError(
                        f"dlg_list_param {attr_name} is not a list. Type is {type(value)}"
                    )
            elif isinstance(member, dlg_dict_param):
                value = get_param_value(attr_name, member.default_value)
                if isinstance(value, str):
                    if value == "":
                        value = {}
                    else:
                        value = ast.literal_eval(value)
                if value is not None and not isinstance(value, dict):
                    raise TypeError(
                        "dlg_dict_param {} is not a dict. It is a {}".format(
                            attr_name, type(value)
                        )
                    )
            else:
                continue
            setattr(self, attr_name, value)

    def _popArg(self, kwargs, key, default):
        """
        Pops the specified key arg from kwargs else returns the default
        """
        if key not in kwargs:
            logger.debug("Defaulting %s to %s in %r", key, str(default), self)
        return kwargs.pop(key, default)

    def __hash__(self):
        return hash(self._uid)

    def __repr__(self):
        return "<%s oid=%s, uid=%s>" % (
            self.__class__.__name__,
            self.oid,
            self.uid,
        )

    def initialize(self, **kwargs):
        """
        Performs any specific subclass initialization.

        `kwargs` contains all the keyword arguments given at construction time,
        except those used by the constructor itself. Implementations of this
        method should make sure that arguments in the `kwargs` dictionary are
        removed once they are interpreted so they are not interpreted by
        accident by another method implementations that might reside in the call
        hierarchy (in the case that a subclass implementation calls the parent
        method implementation, which is usually the case).
        """

    def autofill_environment_variables(self):
        """
        Runs through all parameters here, fetching those which match the env-var syntax when
        discovered.
        """
        for param_key, param_val in self.parameters.items():
            if self._env_var_matcher.fullmatch(str(param_val)):
                self.parameters[param_key] = self.get_environment_variable(param_val)
            if self._dlg_var_matcher.fullmatch(str(param_val)):
                self.parameters[param_key] = getDlgVariable(param_val)

    def get_environment_variable(self, key: str):
        """
        Expects keys of the form $store_name.var_name
        $store_name.var_name.sub_var_name will query store_name for var_name.sub_var_name
        """
        if self._dlg_var_matcher.fullmatch(key):
            return getDlgVariable(key)
        if len(key) < 2 or key[0] != "$":
            # Reject malformed entries
            return key
        key_edit = key[1:]
        env_var_ref, env_var_key = (
            key_edit.split(".")[0],
            ".".join(key_edit.split(".")[1:]),
        )
        env_var_drop = None
        for producer in self._producers:
            if producer.name == env_var_ref:
                env_var_drop = producer
        if env_var_drop is not None:  # TODO: Check for KeyValueDROP interface support
            ret_val = env_var_drop.get(env_var_key)
            if ret_val is None:
                return key
            return ret_val
        else:
            return key

    def get_environment_variables(self, keys: list):
        """
        Expects multiple instances of the single key form
        """
        return_values = []
        for key in keys:
            # TODO: Accumulate calls to the same env_var_store to save communication
            return_values.append(self.get_environment_variable(key))
        return return_values

    @property
    def merkleroot(self):
        return self._merkleRoot

    @property
    def reproducibility_level(self):
        return self._reproducibility

    @reproducibility_level.setter
    def reproducibility_level(self, new_flag):
        if type(new_flag) != ReproducibilityFlags:
            raise TypeError("new_flag must be a reproducibility flag enum.")
        elif rmode_supported(new_flag):  # TODO: Support custom checkers for repro-level
            self._reproducibility = new_flag
            if new_flag == ReproducibilityFlags.ALL:
                self._committed = False
                self._merkleRoot = {rmode.name: None for rmode in ALL_RMODES}
                self._merkleTree = {rmode.name: None for rmode in ALL_RMODES}
                self._merkleData = {rmode.name: [] for rmode in ALL_RMODES}
            elif self._committed:
                # Current behaviour, set to un-committed again after change
                self._committed = False
                self._merkleRoot = None
                self._merkleTree = None
                self._merkleData = []
        else:
            raise NotImplementedError("new_flag %d is not supported" % new_flag.value)

    def generate_rerun_data(self):
        """
        Provides a serailized list of Rerun data.
        At runtime, Rerunning only requires execution success or failure.
        :return: A dictionary containing rerun values
        """
        return {"status": self._status}

    def generate_repeat_data(self):
        """
        Provides a list of Repeat data.
        At runtime, repeating, like rerunning only requires execution success or failure.
        :return: A dictionary containing runtime exclusive repetition values.
        """
        return {"status": self._status}

    def generate_recompute_data(self):
        """
        Provides a dictionary containing recompute data.
        At runtime, recomputing, like repeating and rerunning, by default, only shows success or failure.
        We anticipate that any further implemented behaviour be done at a lower class.
        :return: A dictionary containing runtime exclusive recompute values.
        """
        return {"status": self._status}

    def generate_reproduce_data(self):
        """
        Provides a list of Reproducibility data (specifically).
        The default behaviour is to return nothing. Per-class behaviour is to be achieved by overriding this method.
        :return: A dictionary containing runtime exclusive reproducibility data.
        """
        return {}

    def generate_replicate_sci_data(self):
        """
        Provides a list of scientific replication data.
        This is by definition a merging of both reproduction and rerun data
        :return: A dictionary containing runtime exclusive scientific replication data.
        """
        res = {}
        res.update(self.generate_rerun_data())
        res.update(self.generate_reproduce_data())
        return res

    def generate_replicate_comp_data(self):
        """
        Provides a list of computational replication data.
        This is by definition a merging of both reproduction and recompute data
        :return: A dictionary containing runtime exclusive computational replication data.
        """
        res = {}
        recomp_data = self.generate_recompute_data()
        if recomp_data is not None:
            res.update(self.generate_recompute_data())
        res.update(self.generate_reproduce_data())
        return res

    def generate_replicate_total_data(self):
        """
        Provides a list of total replication data.
        This is by definition a merging of reproduction and repetition data
        :return: A dictionary containing runtime exclusive total replication data.
        """
        res = {}
        res.update(self.generate_repeat_data())
        res.update(self.generate_reproduce_data())
        return res

    def generate_merkle_data(self):
        """
        Provides a serialized summary of data as a list.
        Fields constitute a single entry in this list.
        Wraps several methods dependent on this DROPs reproducibility level
        Some of these are abstract.
        :return: A dictionary of elements constituting a summary of this drop
        """
        if self._reproducibility is ReproducibilityFlags.NOTHING:
            return {}
        elif self._reproducibility is ReproducibilityFlags.RERUN:
            return self.generate_rerun_data()
        elif self._reproducibility is ReproducibilityFlags.REPEAT:
            return self.generate_repeat_data()
        elif self._reproducibility is ReproducibilityFlags.RECOMPUTE:
            return self.generate_recompute_data()
        elif self._reproducibility is ReproducibilityFlags.REPRODUCE:
            return self.generate_reproduce_data()
        elif self._reproducibility is ReproducibilityFlags.REPLICATE_SCI:
            return self.generate_replicate_sci_data()
        elif self._reproducibility is ReproducibilityFlags.REPLICATE_COMP:
            return self.generate_replicate_comp_data()
        elif self._reproducibility is ReproducibilityFlags.REPLICATE_TOTAL:
            return self.generate_replicate_total_data()
        elif self._reproducibility is ReproducibilityFlags.ALL:
            return {
                ReproducibilityFlags.RERUN.name: self.generate_rerun_data(),
                ReproducibilityFlags.REPEAT.name: self.generate_repeat_data(),
                ReproducibilityFlags.RECOMPUTE.name: self.generate_recompute_data(),
                ReproducibilityFlags.REPRODUCE.name: self.generate_reproduce_data(),
                ReproducibilityFlags.REPLICATE_SCI.name: self.generate_replicate_sci_data(),
                ReproducibilityFlags.REPLICATE_COMP.name: self.generate_replicate_comp_data(),
                ReproducibilityFlags.REPLICATE_TOTAL.name: self.generate_replicate_total_data(),
            }
        else:
            raise NotImplementedError("Currently other levels are not in development.")

    def commit(self):
        """
        Generates the MerkleRoot of this DROP
        Should only be called once this DROP is completed.
        """
        if not self._committed:
            #  Generate the MerkleData
            self._merkleData = self.generate_merkle_data()
            if self._reproducibility == ReproducibilityFlags.ALL:
                for rmode in ALL_RMODES:
                    self._merkleTree[rmode.name] = MerkleTree(
                        self._merkleData[rmode.name].items(), common_hash
                    )
                    self._merkleRoot[rmode.name] = self._merkleTree[
                        rmode.name
                    ].merkle_root
            else:
                # Fill MerkleTree, add data and set the MerkleRoot Value
                self._merkleTree = MerkleTree(self._merkleData.items(), common_hash)
                self._merkleRoot = self._merkleTree.merkle_root
                # Set as committed
            self._committed = True
        else:
            logger.debug("Trying to re-commit DROP %s, cannot overwrite.", self)

    @property
    def oid(self):
        """
        The DROP's Object ID (OID). OIDs are unique identifiers given to
        semantically different DROPs (and by consequence the data they
        represent). This means that different DROPs that point to the same
        data semantically speaking, either in the same or in a different
        storage, will share the same OID.
        """
        return self._oid

    @property
    def uid(self):
        """
        The DROP's Unique ID (UID). Unlike the OID, the UID is globally
        different for all DROP instances, regardless of the data they
        point to.
        """
        return self._uid

    @property
    def type(self):
        return self._type

    @property
    def executionMode(self):
        """
        The execution mode of this DROP. If `ExecutionMode.DROP` it means
        that this DROP will automatically trigger the execution of all its
        consumers. If `ExecutionMode.EXTERNAL` it means that this DROP
        will *not* trigger its consumers, and therefore an external entity will
        have to do it.
        """
        return self._executionMode

    def handleInterest(self, drop):
        """
        Main mechanism through which a DROP handles its interest in a
        second DROP it isn't directly related to.

        A call to this method should be expected for each DROP this
        DROP is interested in. The default implementation does nothing,
        but implementations are free to perform any action, such as subscribing
        to events or storing information.

        At this layer only the handling of such an interest exists. The
        expression of such interest, and the invocation of this method wherever
        necessary, is currently left as a responsibility of the entity creating
        the DROPs. In the case of a Session in a DROPManager for
        example this step would be performed using deployment-time information
        contained in the dropspec dictionaries held in the session.
        """

    def _fire(self, eventType: str, **kwargs):
        """
        Delivers an event of `eventType` to all interested listeners.

        All the key-value pairs contained in `attrs` are set as attributes of
        the event being sent. On top of that, the `uid` and `oid` attributes are
        also added, carrying the uid and oid of the current DROP, respectively.
        """
        kwargs["oid"] = self.oid
        kwargs["uid"] = self.uid
        kwargs["session_id"] = self._dlg_session_id
        kwargs["name"] = self.name
        kwargs["lg_key"] = self.lg_key
        self._fireEvent(eventType, **kwargs)

    @property
    def phase(self):
        """
        This DROP's phase. The phase indicates the resilience of a DROP.
        """
        return self._phase

    @phase.setter
    def phase(self, phase):
        self._phase = phase

    @property
    def targetPhase(self):
        return self._targetPhase

    @property
    def expirationDate(self):
        return self._expirationDate

    @property
    def expireAfterUse(self):
        return self._expireAfterUse

    @property
    def size(self):
        """
        The size of the data pointed by this DROP. Its value is
        automatically calculated if the data was actually written through this
        DROP (using the `self.write()` method directly or indirectly). In
        the case that the data has been externally written, the size can be set
        externally after the DROP has been moved to COMPLETED or beyond.
        """
        return self._size

    @size.setter
    def size(self, size):
        if self._size is not None:
            raise RuntimeError(
                "The size of DROP %s is already calculated, cannot overwrite with new value"
                % (self)
            )
        if self.status in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            raise RuntimeError(
                "DROP %s is still not fully written, cannot manually set a size yet"
                % (self)
            )
        self._size = size

    @property
    def persist(self):
        """
        Whether this DROP should be considered persisted after completion
        """
        return self._persist

    @property
    def status(self):
        """
        The current status of this DROP.
        """
        with self._statusLock:
            return self._status

    @status.setter
    def status(self, value):
        with self._statusLock:
            # if we are already in the state that is requested then do nothing
            if value == self._status:
                return
            self._status = value

        self._fire("status", status=value)

    @property
    def parent(self):
        """
        The DROP that acts as the parent of the current one. This
        parent/child relationship is created by ContainerDROPs, which are
        a specific kind of DROP.
        """
        return self._parent

    @parent.setter
    @track_current_drop
    def parent(self, parent):
        if self._parent and parent:
            logger.warning(
                "A parent is already set in %r, overwriting with new value",
                self.oid,
            )
        if parent:
            prevParent = self._parent
            self._parent = parent  # a parent is a container
            if hasattr(parent, "addChild") and self not in parent.children:
                try:
                    parent.addChild(self)
                except InvalidRelationshipException:
                    self._parent = prevParent

    def get_consumers_nodes(self):
        """
        Gets the physical node address(s) of the consumer of this drop.
        """
        return [cons.node for cons in self._consumers] + [
            cons.node for cons in self._streamingConsumers
        ]

    @property
    def consumers(self):
        """
        The list of 'normal' consumers held by this DROP.

        :see: `self.addConsumer()`
        """
        return self._consumers[:]

    @track_current_drop
    def addConsumer(self, consumer, back=True):
        """
        Adds a consumer to this DROP.

        Consumers are normally (but not necessarily) AppDROPs that get
        notified when this DROP moves into the COMPLETED or ERROR states.
        This is done by firing an event of type `dropCompleted` to which the
        consumer subscribes to.

        This is one of the key mechanisms by which the DROP graph is
        executed automatically. If AppDROP B consumes DROP A, then
        as soon as A transitions to COMPLETED B will be notified and will
        probably start its execution.
        """

        # An object cannot be a normal and streaming consumer at the same time,
        # see the comment in the __init__ method
        cuid = consumer.uid
        if cuid in self._streamingConsumers_uids:
            raise InvalidRelationshipException(
                DROPRel(consumer, DROPLinkType.CONSUMER, self),
                "Consumer already registered as a streaming consumer",
            )

        # Add if not already present
        # Add the reverse reference too automatically
        if cuid in self._consumers_uids:
            return
        # logger.debug("Adding new consumer %r to %r", consumer.oid, self.oid)
        self._consumers.append(consumer)

        # Subscribe the consumer to events sent when this DROP moves to
        # COMPLETED. This way the consumer will be notified that its input has
        # finished.
        # This only happens if this DROP's execution mode is 'DROP'; otherwise
        # an external entity will trigger the execution of the consumer at the
        # right time
        if self.executionMode == ExecutionMode.DROP:
            self.subscribe(consumer, "dropCompleted")

        # Automatic back-reference
        if back and hasattr(consumer, "addInput"):
            logger.debug("Adding back %r as input of %r", self.oid, consumer)
            consumer.addInput(self, False)

        # Add reproducibility subscription
        self.subscribe(consumer, "reproducibility")

    @property
    def producers(self):
        """
        The list of producers that write to this DROP

        :see: `self.addProducer()`
        """
        return self._producers[:]

    @track_current_drop
    def addProducer(self, producer, back=True):
        """
        Adds a producer to this DROP.

        Producers are AppDROPs that write into this DROP; from the
        producers' point of view, this DROP is one of its many outputs.

        When a producer has finished its execution, this DROP will be
        notified via the self.producerFinished() method.
        """

        # Don't add twice
        puid = producer.uid
        if puid in self._producers_uids:
            return

        self._producers.append(producer)

        # Automatic back-reference
        if back and hasattr(producer, "addOutput"):
            producer.addOutput(self, False)

    @track_current_drop
    def handleEvent(self, e):
        """
        Handles the arrival of a new event. Events are delivered from those
        objects this DROP is subscribed to.
        """
        if e.type == "producerFinished":
            self.producerFinished(e.uid, e.status)
        elif e.type == "reproducibility":
            self.dropReproComplete(e.uid, e.reprodata)

    @track_current_drop
    def producerFinished(self, uid, drop_state):
        """
        Method called each time one of the producers of this DROP finishes
        its execution. Once all producers have finished this DROP moves to the
        COMPLETED state (or to ERROR if one of the producers is on the ERROR
        state).

        This is one of the key mechanisms through which the execution of a
        DROP graph is accomplished. If AppDROP A produces DROP
        B, as soon as A finishes its execution B will be notified and will move
        itself to COMPLETED.
        """

        finished = False
        with self._finishedProducersLock:
            self._finishedProducers.append(drop_state)
            nFinished = len(self._finishedProducers)
            nProd = len(self._producers)
            self._refCount -= 1

            if nFinished > nProd:
                raise RuntimeError(
                    "More producers finished that registered in DROP %r: %d > %d"
                    % (self, nFinished, nProd)
                )
            elif nFinished == nProd:
                finished = True

        if finished:
            logger.debug("All producers finished for DROP %r,%s", self, uid)

            # decided that if any producer fails then fail the data drop
            if DROPStates.ERROR in self._finishedProducers:
                self.setError()
            else:
                self.setCompleted()

    def dropReproComplete(self, uid, reprodata):
        """
        Callback invoved when a DROP with UID `uid` has finishing processing its reproducibility information.
        Importantly, this is independent of that drop being completed.
        """
        #  TODO: Perform some action

    @property
    def streamingConsumers(self):
        """
        The list of 'streaming' consumers held by this DROP.

        :see: `self.addStreamingConsumer()`
        """
        return self._streamingConsumers[:]

    @track_current_drop
    def addStreamingConsumer(self, streamingConsumer, back=True):
        """
        Adds a streaming consumer to this DROP.

        Streaming consumers are AppDROPs that receive the data written
        into this DROP *as it gets written*, and therefore do not need to
        wait until this DROP has been moved to the COMPLETED state.
        """

        # An object cannot be a normal and streaming streamingConsumer at the same time,
        # see the comment in the __init__ method
        scuid = streamingConsumer.uid
        if scuid in self._consumers_uids:
            raise InvalidRelationshipException(
                DROPRel(streamingConsumer, DROPLinkType.STREAMING_CONSUMER, self),
                "Consumer is already registered as a normal consumer",
            )

        # Add if not already present
        if scuid in self._streamingConsumers_uids:
            return
        logger.debug(
            "Adding new streaming streaming consumer for %r: %s",
            self,
            streamingConsumer,
        )
        self._streamingConsumers.append(streamingConsumer)

        # Automatic back-reference
        if back and hasattr(streamingConsumer, "addStreamingInput"):
            streamingConsumer.addStreamingInput(self, False)

        # Subscribe the streaming consumer to events sent when this DROP moves
        # to COMPLETED. This way the streaming consumer will be notified that
        # its input has finished
        # This only happens if this DROP's execution mode is 'DROP'; otherwise
        # an external entity will trigger the execution of the consumer at the
        # right time
        if self.executionMode == ExecutionMode.DROP:
            self.subscribe(streamingConsumer, "dropCompleted")

        # Add reproducibility subscription
        self.subscribe(streamingConsumer, "reproducibility")

    def completedrop(self):
        """
        Builds final reproducibility data for this drop and fires a 'dropComplete' event.
        This should be called once a drop is finished in success or error
        :return:
        """
        self.commit()
        reprodata = {"data": self._merkleData, "merkleroot": self.merkleroot}
        self._fire(eventType="reproducibility", reprodata=reprodata)
        if logger.getEffectiveLevel() != logging.getLevelName(self._global_log_level):
            logger.warning(
                "log-level after %s.%s: %s",
                self.name,
                self._humanKey,
                logging.getLevelName(logger.getEffectiveLevel()),
            )

    @track_current_drop
    def setError(self):
        """
        Moves this DROP to the ERROR state.
        """

        if self.status in (DROPStates.CANCELLED, DROPStates.SKIPPED):
            return

        self._closeWriters()

        logger.info("Moving %r to ERROR", self)
        self.status = DROPStates.ERROR

        # Signal our subscribers that the show is over
        self._fire(eventType="dropCompleted", status=DROPStates.ERROR)
        self.completedrop()

    def _setCompletedStateCheck(self) -> bool:
        """
        Checks DROP state to identify conflics before setting it to completed
        """
        status = self.status
        if status == DROPStates.CANCELLED:
            return False
        elif status == DROPStates.SKIPPED:
            self._fire("dropCompleted", status=status)
            return False
        elif status == DROPStates.COMPLETED:
            logger.warning("%r already in COMPLETED state", self)
            return False
        elif status not in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            raise Exception(
                "%r not in INITIALIZED or WRITING state (%s), cannot setComplete()"
                % (self, self.status)
            )
        return True

    @track_current_drop
    def setCompleted(self):
        """
        Moves this DROP to the COMPLETED state. This can be used when not all the
        expected data has arrived for a given DROP, but it should still be moved
        to COMPLETED, or when the expected amount of data held by a DROP
        is not known in advanced.
        """
        if not self._setCompletedStateCheck():
            return
        try:
            self._closeWriters()
        except AttributeError as exp:
            logger.debug(exp)

        logger.debug("Moving %r to COMPLETED", self)
        self.status = DROPStates.COMPLETED
        # Signal our subscribers that the show is over
        self._fire(eventType="dropCompleted", status=DROPStates.COMPLETED)
        self.completedrop()

    def isCompleted(self):
        """
        Checks whether this DROP is currently in the COMPLETED state or not
        """
        # Mind you we're not accessing _status, but status. This way we use the
        # lock in status() to access _status
        return self.status == DROPStates.COMPLETED

    def cancel(self):
        """Moves this drop to the CANCELLED state closing any writers we opened"""
        if self.status in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            self._closeWriters()
            self.status = DROPStates.CANCELLED
        self.completedrop()

    def skip(self):
        """
        Moves this drop to the SKIPPED state closing any writers we opened.

        If the drop has more than one producer it will block until the first
        producer is skipped or all producers are completed. If the block_skip
        flag is set the drop will block a skipping chain until the last producer
        is completed or also skipped. This is useful for branches inside loops
        to allow alternate paths for each iteration. For a branch to terminate
        a loop the flag needs to be false (default).
        """
        if self.status in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            if not self.block_skip or (
                    len(self.producers) == 1
                    or (len(self.producers) > 1 and self._refCount == 1)
            ):
                self._closeWriters()
                self.status = DROPStates.SKIPPED
                self.completedrop()
            else:
                self._refCount -= 1

    @property
    def node(self):
        return self._node

    @property
    def dataIsland(self):
        return self._dataIsland

    @property
    def parameters(self):
        return self._parameters

    @property
    def dlg_session_id(self) -> str:
        """
        Get the session id for the session to which this DROP is assigned.
        """
        return self._dlg_session_id

    @property
    def humanKey(self) -> str:
        """
        Get the Human Readable Key for this DROP
        """
        return self._humanKey

    @property
    def log_level(self):
        """
        Get the log level for this DROP
        """
        return self._log_level


# Dictionary mapping 1-to-many DROPLinkType constants to the corresponding methods
# used to append a a DROP into a relationship collection of another
# (e.g., one uses `addConsumer` to add a DROPLinkeType.CONSUMER DROP into
# another)
LINKTYPE_1TON_APPEND_METHOD = {
    DROPLinkType.CONSUMER: "addConsumer",
    DROPLinkType.STREAMING_CONSUMER: "addStreamingConsumer",
    DROPLinkType.INPUT: "addInput",
    DROPLinkType.STREAMING_INPUT: "addStreamingInput",
    DROPLinkType.OUTPUT: "addOutput",
    DROPLinkType.CHILD: "addChild",
    DROPLinkType.PRODUCER: "addProducer",
}

# Same as above, but for N-to-1 relationships, in which case we indicate not a
# method but a property
LINKTYPE_NTO1_PROPERTY = {DROPLinkType.PARENT: "parent"}

LINKTYPE_1TON_BACK_APPEND_METHOD = {
    DROPLinkType.CONSUMER: "addInput",
    DROPLinkType.STREAMING_CONSUMER: "addStreamingInput",
    DROPLinkType.INPUT: "addConsumer",
    DROPLinkType.STREAMING_INPUT: "addStreamingConsumer",
    DROPLinkType.OUTPUT: "addProducer",
    DROPLinkType.CHILD: "setParent",
    DROPLinkType.PRODUCER: "addOutput",
}

LINKTYPE_NTO1_BACK_APPEND_METHOD = {DROPLinkType.PARENT: "addChild"}



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import logging
from dlg.apps.app_base import BarrierAppDROP
from dlg.data.drops.container import ContainerDROP
from ..droputils import DROPFile
from dlg.data.io import NgasIO, OpenMode, NgasLiteIO
from ..meta import (
    dlg_string_param,
    dlg_int_param,
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
)

logger = logging.getLogger(f"dlg.{__name__}")


class ExternalStoreApp(BarrierAppDROP):
    """
    An application that takes its input DROP (which must be one, and only
    one) and creates a copy of it in a completely external store, from the point
    of view of the DALiuGE framework.

    Because this application copies the data to an external location, it also
    shouldn't contain any output, making it a leaf node of the physical graph
    where it resides.
    """

    component_meta = dlg_component(
        "ExternalStoreApp",
        "An application that takes its input DROP (which must be one, and only one) "
        "and creates a copy of it in a completely external store, from the point "
        "of view of the DALiuGE framework.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    def run(self):
        # Check that the constrains are correct
        if self.outputs:
            raise Exception("No outputs should be declared for this application")
        if len(self.inputs) != 1:
            raise Exception("Only one input is expected by this application")

        # ... and go!
        inDrop = self.inputs[0]
        self.store(inDrop)

    def store(self, inputDrop):
        """
        Method implemented by subclasses. It should stores the contents of
        `inputDrop` into an external store.
        """


##
# @brief NgasArchivingApp
# @details Takes an input and archives it in an NGAS server.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param ngasSrv localhost/String/ApplicationArgument/NoPort/ReadWrite//False/False/URL of the NGAS Server
# @param ngasPort 7777/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/"TCP/IP Port on the NGAS Server"
# @param ngasMime "application/octet-stream"/String/ApplicationArgument/NoPort/ReadWrite//False/False/Mime-type of the NGAS payload
# @param ngasTimeout 2/Integer/ApplicationArgument/NoPort/ReadOnly//False/False/Archiving request timeout
# @param ngasConnectTimeout 2/Integer/ApplicationArgument/NoPort/ReadOnly//False/False/NGAS Server connection timeout
# @param fileObject /Object.File/ApplicationArgument/InputPort/ReadWrite//False/False/Input File Object
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.archiving.NgasArchivingApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name archiving/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/the allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @par EAGLE_END
class NgasArchivingApp(ExternalStoreApp):
    """
    An ExternalStoreApp class that takes its input DROP and archives it in
    an NGAS server. It currently deals with non-container DROPs only.

    The archiving to NGAS occurs through the framework and not by spawning a
    new NGAS client process. This way we can read the different storage types
    supported by the framework, and not only filesystem objects.
    """

    component_meta = dlg_component(
        "NgasArchivingApp",
        "An ExternalStoreApp class that takes its input DROP and archives it in "
        "an NGAS server. It currently deals with non-container DROPs only.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    ngasSrv = dlg_string_param("ngasSrv", "localhost")
    ngasPort = dlg_int_param("ngasPort", 7777)
    ngasMime = dlg_string_param("ngasMime", "application/octet-stream")
    ngasTimeout = dlg_int_param("ngasTimeout", 2)
    ngasConnectTimeout = dlg_int_param("ngasConnectTimeout", 2)


    def store(self, inputDrop):
        logger.debug("NGAS Server %s", self.ngasSrv)
        logger.debug("NGAS Port %s", self.ngasPort)

        if isinstance(inputDrop, ContainerDROP):
            raise Exception(
                "ContainerDROPs are not supported as inputs for this application"
            )

        if inputDrop.size is None or inputDrop.size < 0:
            logger.error(
                "NGAS requires content-length to be know, but the given input does not provide a size."
            )
            size = None
        else:
            size = inputDrop.size
            logger.debug("Content-length %s", size)
        try:
            ngasIO = NgasIO(
                self.ngasSrv,
                inputDrop.uid,
                self.ngasPort,
                self.ngasConnectTimeout,
                self.ngasTimeout,
                size,
                mimeType=self.ngasMime,
            )
        except ImportError:
            logger.warning("NgasIO library not available, falling back to NgasLiteIO.")
            ngasIO = NgasLiteIO(
                self.ngasSrv,
                inputDrop.uid,
                self.ngasPort,
                self.ngasConnectTimeout,
                self.ngasTimeout,
                size,
                mimeType=self.ngasMime,
            )

        ngasIO.open(OpenMode.OPEN_WRITE)

        # Copy in blocks of 4096 bytes
        with DROPFile(inputDrop) as f:
            while True:
                buff = f.read(4096)
                ngasIO.write(buff)
                if len(buff) != 4096:
                    break
        ngasIO.close()



from dlg.apps.app_base import BarrierAppDROP


##
# @brief Scatter
# @details A Scatter template drop
# @par EAGLE_START
# @param category Scatter
# @param tag template
# @param num_of_copies 4/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Specifies the number of replications of the content of the scatter construct
# @par EAGLE_END
class ScatterDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a Scatter in the template palette
    """


##
# @brief Gather
# @details A Gather template drop
# @par EAGLE_START
# @param category Gather
# @param tag template
# @param num_of_inputs 2/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Number of inputs
# @param gather_axis 0/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/Index of gather axis
# @par EAGLE_END
class GatherDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a GroupBy in the template palette
    """


##
# @brief Loop
# @details A loop template drop
# @par EAGLE_START
# @param category Loop
# @param tag template
# @param num_of_iter 2/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Number of iterations
# @par EAGLE_END
class LoopDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a loop in the template palette
    """


##
# @brief MKN
# @details A MKN template drop
# @par EAGLE_START
# @param category MKN
# @param tag template
# @param k 1/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Internal multiplicity
# @par EAGLE_END
class MKNDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a MKN in the template palette
    """


##
# @brief GroupBy
# @details A GroupBy template drop
# @par EAGLE_START
# @param category GroupBy
# @param tag template
# @param num_of_inputs 2/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Number of inputs
# @param gather_axis 0/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/Index of gather axis
# @par EAGLE_END
class GroupByDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a GroupBy in the template palette
    """


##
# @brief SubGraph
# @details A SubGraph template drop
# @par EAGLE_START
# @param category SubGraph
# @param tag template
# @par EAGLE_END
class SubGraphDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a SubGraph in the template palette
    """


##
# @brief Comment
# @details A comment template drop
# @par EAGLE_START
# @param category Comment
# @param dropclass dlg.apps.constructs.CommentDrop/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name constructs/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param tag template
# @par EAGLE_END
class CommentDrop(BarrierAppDROP):
    """
    This only exists to make sure we have a comment in the template palette
    """


##
# @brief Exclusive Force Node
# @details An Exclusive Force Node
# @par EAGLE_START
# @param category ExclusiveForceNode
# @param dropclass dlg.apps.constructs.ExclusiveForceDrop/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name construct/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param tag template
# @par EAGLE_END
class ExclusiveForceDrop(BarrierAppDROP):
    """
    This only exists to make sure we have an exclusive force node in the template palette
    """



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite `3`30, Boston,
#    MA 02111-1307  USA
#
"""
Start the DALiuGE cluster on various facilities

Current plan (as of 12-April-2016):
    1. Launch a number of Node Managers (NM) using MPI processes
    2. Having the NM MPI processes to send their IP addresses to the Rank 0
       MPI process
    3. Launch the Island Manager (IM) on the Rank 0 MPI process using those IP
       addresses
"""

import json
import logging
import multiprocessing
import optparse # pylint: disable=deprecated-module
import os
import socket
import subprocess
import sys
import threading
import time
import uuid

from dlg.deploy import dlg_proxy, remotes
from dlg.deploy import common
from dlg import utils
from dlg.common import tool
from dlg.dropmake import pg_generator
from dlg.translator.tool_commands import parse_partition_algo_params
from dlg.manager import cmdline
from dlg.manager.client import NodeManagerClient
from dlg.constants import (
    NODE_DEFAULT_REST_PORT,
    ISLAND_DEFAULT_REST_PORT,
    MASTER_DEFAULT_REST_PORT,
)
from dlg.common.reproducibility.reproducibility import (
    init_pgt_unroll_repro_data,
    init_pgt_partition_repro_data,
    init_pg_repro_data,
)

DIM_WAIT_TIME = 60
MM_WAIT_TIME = DIM_WAIT_TIME
GRAPH_SUBMIT_WAIT_TIME = 10
GRAPH_MONITOR_INTERVAL = 5
VERBOSITY = "5"
LOGGER = logging.getLogger("deploy.dlg.cluster")
APPS = (
    None,
    "test.graphsRepository.SleepApp",
    "test.graphsRepository.SleepAndCopyApp",
)


def check_host(host, port: int, timeout: int = 5, check_with_session=False):
    """
    Checks if a given host/port is up and running (i.e., it is open).
    If ``check_with_session`` is ``True`` then it is assumed that the
    host/port combination corresponds to a Node Manager and the check is
    performed by attempting to create and delete a session.
    """
    if not check_with_session:
        return utils.portIsOpen(host, port, timeout)

    try:
        session_id = str(uuid.uuid4())
        with NodeManagerClient(host, port, timeout=timeout) as client:
            client.create_session(session_id)
            client.destroy_session(session_id)
        return True
    except Exception: #pylint: disable=broad-exception-caught
        return False


def check_hosts(
    ips,
    timeout=None,
    check_with_session=False,
    retry=1,
):
    """
    Check that the given list of IPs are all up in the given port within the
    given timeout, and returns the list of IPs that were found to be up.
    """

    def check_and_add(ip_addr):
        ntries = retry
        if ":" in ip_addr:
            ip, port = ip_addr.split(":")
            port = int(port)
        else:
            ip = ip_addr
            port = NODE_DEFAULT_REST_PORT
        while ntries:
            if check_host(
                ip,
                port,
                timeout=timeout,
                check_with_session=check_with_session,
            ):
                LOGGER.info("Host %s:%d is running", ip, port)
                return f"{ip}:{port}"
            LOGGER.warning("Failed to contact host %s:%d", ip, port)
            ntries -= 1
        return None

    # Don't return None values
    thread_pool = multiprocessing.pool.ThreadPool(min(50, len(ips)))
    result_pool = thread_pool.map(check_and_add, ips)
    thread_pool.close()
    thread_pool.join()

    return [f"{ip}" for ip in result_pool if ip]


def get_ip_via_ifconfig(iface_index):
    out = subprocess.check_output("ifconfig")
    ifaces_info = list(filter(None, out.split(b"\n\n")))
    LOGGER.info("Found %d interfaces, getting %d", len(ifaces_info), iface_index)
    for line in ifaces_info[iface_index].splitlines():
        line = line.strip()
        if line.startswith(b"inet"):
            return utils.b2s(line.split()[1])
    raise ValueError("Interace %d is not an IP interface" % iface_index)


def get_ip_via_netifaces(iface_index):
    return utils.get_local_ip_addr()[iface_index][0]


def start_node_mgr(
    log_dir,
    my_ip,
    logv=1,
    max_threads=0,
    host=None,
    event_listeners="",
    use_tool=True,
):
    """
    Start node manager
    """
    LOGGER.info("Starting node manager on host %s", my_ip)
    host = host or "0.0.0.0"
    log_level = "v" * logv
    args = [
        "-l",
        log_dir,
        "-%s" % log_level,
        "-H",
        host,
        "-m",
        "1024",
        "-t",
        str(max_threads),
    ]
    if event_listeners:
        args += ["--event-listeners", event_listeners]

    if use_tool:
        # This returns immediately
        # proc = tool.start_process("nm", args)
        proc = tool.start_process("nm", args)
        LOGGER.info("Node manager process started with pid %d", proc.pid)
        return proc
    else:
        # This blocks until NM shutdown externally
        return cmdline.dlgNM(optparse.OptionParser(), args)


def start_dim(node_list, log_dir, origin_ip, logv=1):
    """
    Start data island manager
    """
    LOGGER.info(
        "Starting island manager on host %s for node managers %r",
        origin_ip,
        node_list,
    )
    log_level = "v" * logv
    args = [
        "-l",
        log_dir,
        "-%s" % log_level,
        "-N",
        ",".join(node_list),
        "-H",
        "0.0.0.0",
        "-m",
        "2048",
        "--dump_graphs"
    ]
    proc = tool.start_process("dim", args)
    LOGGER.info("Island manager process started with pid %d", proc.pid)
    return proc


def start_mm(node_list, log_dir, logv=1):
    """
    Start master manager

    node_list:  a list of node address that host DIMs
    """
    log_level = "v" * logv
    args = [
        "-l",
        log_dir,
        "-N",
        ",".join(node_list),
        "-%s" % log_level,
        "-H",
        "0.0.0.0",
        "-m",
        "2048",
        "--dump_graphs"
    ]
    proc = tool.start_process("mm", args)
    LOGGER.info("Master manager process started with pid %d", proc.pid)
    return proc


#    cmdline.dlgMM(parser, args)


def _stop(endpoints):
    LOGGER.info("Stopping ThreadPool")

    def _the_stop(endpoint):
        common.BaseDROPManagerClient(endpoint[0], endpoint[1]).stop()

    thread_pool = multiprocessing.pool.ThreadPool(min(50, len(endpoints)))
    thread_pool.map(_the_stop, endpoints)
    thread_pool.close()
    thread_pool.join()


def stop_nms(ips):
    LOGGER.info("Stopping node managers on nodes % s", ips)
    _stop([(ip, NODE_DEFAULT_REST_PORT) for ip in ips])


def stop_dims(ips):
    LOGGER.info("Stopping island managers on nodes %s", ips)
    _stop([(ip, ISLAND_DEFAULT_REST_PORT) for ip in ips])


def stop_mm(ip_addr):
    LOGGER.info("Stopping master managers on node %s", ip_addr)
    _stop([(ip_addr, MASTER_DEFAULT_REST_PORT)])


def submit_and_monitor(physical_graph, opts, host, port, submit=True):
    def _task():
        dump_path = None
        if opts.dump:
            dump_path = os.path.join(opts.log_dir, "status-monitoring.json")
        if submit:
            session_id = common.submit(
                physical_graph, host=host, port=port, session_id=opts.ssid
            )
        else:
            session_id = opts.ssid

        LOGGER.info("Start monitoring session(s) %s on host %s:%s",
                    session_id, host, port)
        while True:
            try:
                common.monitor_sessions(
                    session_id,
                    host=host,
                    port=port,
                    status_dump_path=dump_path,
                )
                break
            except Exception: #pylint: disable=broad-exception-caught
                LOGGER.exception("Monitoring %s:%s failed, restarting it", host, port)
                time.sleep(5)

    threads = threading.Thread(target=_task)
    threads.start()
    return threads


def start_proxy(dlg_host, dlg_port, monitor_host, monitor_port):
    """
    Start the DALiuGE proxy server
    """
    proxy_id = socket.gethostname() + "%.3f" % time.time()
    server = dlg_proxy.ProxyServer(
        proxy_id, dlg_host, monitor_host, dlg_port, monitor_port
    )
    try:
        server.loop()
    except KeyboardInterrupt:
        LOGGER.warning("Ctrl C - Stopping DALiuGE Proxy server")
        sys.exit(1)
    except Exception: # pylint: disable=broad-exception-caught
        LOGGER.exception("DALiuGE proxy terminated unexpectedly")
        sys.exit(1)


def modify_pg(pgt, modifier):
    parts = modifier.split(",")
    func = utils.get_symbol(parts[0])
    args = list(filter(lambda x: "=" not in x, parts[1:]))
    kwargs = dict(map(lambda x: x.split("="), filter(lambda x: "=" in x, parts[1:])))
    return func(pgt, *args, **kwargs)


def get_pg(opts, nms, dims):
    """Gets the Physical Graph that is eventually submitted to the cluster, if any"""

    if not opts.logical_graph and not opts.physical_graph:
        return []

    num_nms = len(nms)
    num_dims = len(dims)
    if opts.logical_graph:
        unrolled = init_pgt_unroll_repro_data(
            pg_generator.unroll(
                opts.logical_graph, opts.ssid, opts.zerorun, APPS[opts.app]
            )
        )
        reprodata = {}
        if not unrolled[-1].get("oid"):
            reprodata = unrolled.pop()
        algo_params = parse_partition_algo_params(opts.algo_params)
        pgt = pg_generator.partition(
            unrolled,
            opts.part_algo,
            num_partitions=num_nms,
            num_islands=num_dims,
            **algo_params,
        )
        pgt.append(reprodata)
        pgt = init_pgt_partition_repro_data(pgt)
        del unrolled  # quickly dispose of potentially big object
    else:
        with open(opts.physical_graph, "rb") as pg_file:
            pgt = json.load(pg_file)

    # modify the PG as necessary
    for modifier in opts.pg_modifiers.split(":"):
        if modifier is not None and modifier != "":
            modify_pg(pgt, modifier)

    # Check which NMs are up and use only those form now on
    nms = check_hosts(
        nms,
        check_with_session=opts.check_with_session,
        timeout=MM_WAIT_TIME,
        retry=3,
    )
    LOGGER.info("Mapping graph to available resources: nms %s, dims %s",
                nms, dims)
    physical_graph = init_pg_repro_data(
        pg_generator.resource_map(
            pgt, dims + nms, num_islands=num_dims, co_host_dim=opts.co_host_dim
        )
    )
    if opts.physical_graph:
        graph_name = f"{opts.physical_graph.split('.pgt.graph')[0]}.pg.graph"
    else:
        graph_name = os.path.basename(opts.log_dir)
        graph_name = f"{graph_name.split('_')[0]}.pg.graph"
    with open(os.path.join(opts.log_dir, graph_name), "wt") as pg_file:
        json.dump(physical_graph, pg_file)
    return physical_graph


def get_ip(opts):
    find_ip = get_ip_via_ifconfig if opts.use_ifconfig else get_ip_via_netifaces
    return find_ip(opts.interface)


def get_remote(opts):
    my_ip = get_ip(opts)
    if opts.remote_mechanism == "mpi":
        return remotes.MPIRemote(opts, my_ip)
    if opts.remote_mechanism == "dlg":
        return remotes.DALiuGERemote(opts, my_ip)
    if opts.remote_mechanism == "dlg-hybrid":
        return remotes.DALiuGEHybridRemote(opts, my_ip)
    else:  # == 'slurm'
        return remotes.SlurmRemote(opts, my_ip)


def main():
    parser = optparse.OptionParser()
    parser.add_option(
        "-l",
        "--log_dir",
        action="store",
        type="string",
        dest="log_dir",
        help="Log directory (required)",
    )
    # if this parameter is present, it means we want to get monitored
    parser.add_option(
        "-m",
        "--monitor_host",
        action="store",
        type="string",
        dest="monitor_host",
        help="Monitor host IP (optional)",
    )
    parser.add_option(
        "-o",
        "--monitor_port",
        action="store",
        type="int",
        dest="monitor_port",
        help="Monitor port",
        default=dlg_proxy.default_dlg_monitor_port,
    )
    parser.add_option(
        "-v",
        "--verbose-level",
        action="store",
        type="int",
        dest="verbose_level",
        help="Verbosity level (1-3) of the DIM/NM logging",
        default=1,
    )
    parser.add_option(
        "-z",
        "--zerorun",
        action="store_true",
        dest="zerorun",
        help="Generate a physical graph that takes no time to run",
        default=False,
    )
    parser.add_option(
        "--app",
        action="store",
        type="int",
        dest="app",
        help="The app to use in the PG. 1=SleepApp (default), 2=SleepAndCopy",
        default=0,
    )

    parser.add_option(
        "-t",
        "--max-threads",
        action="store",
        type="int",
        dest="max_threads",
        help="Max thread pool size used for executing drops. 0 (default) means no pool.",
        default=0,
    )

    parser.add_option(
        "-L",
        "--logical-graph",
        action="store",
        type="string",
        dest="logical_graph",
        help="The filename of the logical graph to deploy",
        default=None,
    )
    parser.add_option(
        "-P",
        "--physical-graph",
        action="store",
        type="string",
        dest="physical_graph",
        help="The filename of the physical graph (template) to deploy",
        default=None,
    )

    parser.add_option(
        "-s",
        "--num_islands",
        action="store",
        type="int",
        dest="num_islands",
        help="The number of Data Islands",
    )

    parser.add_option(
        "-d",
        "--dump",
        action="store_true",
        dest="dump",
        help="dump file base name?",
        default=False,
    )

    parser.add_option(
        "-i",
        "--interface",
        type="int",
        help="Index of network interface to use as the external interface/address for each host",
        default=0,
    )

    parser.add_option(
        "--part-algo",
        type="string",
        dest="part_algo",
        help="Partition algorithms",
        default="metis",
    )
    parser.add_option(
        "-A",
        "--algo-param",
        action="append",
        dest="algo_params",
        help="Extra name=value parameters used by the algorithms (algorithm-specific)",
        default=[],
    )

    parser.add_option(
        "--ssid", type="string", dest="ssid", help="session id", default=""
    )

    parser.add_option(
        "-u",
        "--all_nics",
        action="store_true",
        dest="all_nics",
        help="Listen on all NICs for a node manager",
        default=True,
    )

    parser.add_option(
        "--check-interfaces",
        action="store_true",
        dest="check_interfaces",
        help="Run a small network interfaces test and exit",
        default=False,
    )
    parser.add_option(
        "--collect-interfaces",
        action="store_true",
        dest="collect_interfaces",
        help="Collect all interfaces and exit",
        default=False,
    )
    parser.add_option(
        "--use-ifconfig",
        action="store_true",
        dest="use_ifconfig",
        help="Use ifconfig to find a suitable external interface/address for each host",
        default=False,
    )
    parser.add_option(
        "-S",
        "--check_with_session",
        action="store_true",
        dest="check_with_session",
        help="Check for node managers' availability by creating/destroy a session",
        default=False,
    )

    parser.add_option(
        "--event-listeners",
        action="store",
        type="string",
        dest="event_listeners",
        help="A colon-separated list of event listener classes to be used",
        default="",
    )

    parser.add_option(
        "--sleep-after-execution",
        action="store",
        type="int",
        dest="sleep_after_execution",
        help="Sleep time interval after graph execution finished",
        default=0,
    )

    parser.add_option(
        "--pg-modifiers",
        help=(
            "A colon-separated list of python functions that modify a PG before submission. "
            "Each specification is in the form of <funcname>[,[arg1=]val1][,[arg2=]val2]..."
        ),
        default="",
    )

    parser.add_option(
        "-r",
        "--remote-mechanism",
        help="The mechanism used by this script to coordinate remote processes",
        choices=["mpi", "slurm", "dlg", "dlg-hybrid"],
        default="mpi",
    )

    parser.add_option(
        "--co-host-dim",
        action="store_true",
        dest="co_host_dim",
        help="Start DIM on first NM node",
    )

    (options, _) = parser.parse_args()

    if options.check_interfaces:
        try:
            print("From netifaces: %s" % get_ip_via_netifaces(options.interface))
        except IndexError:
            LOGGER.exception("Failed to get information via netifaces")
        try:
            print("From ifconfig: %s" % get_ip_via_ifconfig(options.interface))
        except ValueError:
            LOGGER.exception("Failed to get information via ifconfig")
        sys.exit(0)
    elif options.collect_interfaces:
        from mpi4py import MPI

        comm = MPI.COMM_WORLD  # @UndefinedVariable
        ips = comm.allgather(get_ip(options))
        if comm.Get_rank() == 0:
            print(" ".join(ips))
        sys.exit(0)

    if bool(options.logical_graph) == bool(options.physical_graph):
        parser.error(
            "Either a logical graph or physical graph filename must be specified"
        )
    for graph_file_name in (options.logical_graph, options.physical_graph):
        if graph_file_name and not os.path.exists(graph_file_name):
            parser.error("Cannot locate graph file at '{0}'".format(graph_file_name))

    if options.monitor_host is not None and options.num_islands > 1:
        parser.error("We do not support proxy monitor multiple islands yet")

    # if options.ssid == "":
    #     options.ssid = time.

    remote = get_remote(options)

    log_dir = "{0}/{1}".format(options.log_dir, remote.my_ip)
    os.makedirs(log_dir)
    logfile = log_dir + "/start_dlg_cluster.log"
    log_format = (
        "%(asctime)-15s [%(levelname)5.5s] [%(threadName)15.15s] "
        "%(name)s#%(funcName)s:%(lineno)s %(message)s"
    )
    logging.basicConfig(filename=logfile, level=logging.DEBUG, format=log_format)

    LOGGER.info("This node has IP address: %s", remote.my_ip)

    envfile_name = os.path.join(log_dir, "env.txt")
    LOGGER.debug("Dumping process' environment to %s", envfile_name)
    with open(envfile_name, "wt") as env_file:
        for name, value in sorted(os.environ.items()):
            env_file.write("%s=%s\n" % (name, value))

    logv = max(min(3, options.verbose_level), 1)

    # need to dump nodes file first
    if remote.is_highest_level_manager:
        LOGGER.info("Node %s is hosting the highest level manager",
                    remote.my_ip)
        nodesfile = os.path.join(log_dir, "nodes.txt")
        LOGGER.debug("Dumping list of nodes to %s", nodesfile)
        with open(nodesfile, "wt") as env_file:
            env_file.write("\n".join(remote.sorted_peers))
    # start the NM
    LOGGER.debug(
        "my_ip:%s; dim_ips:%s, node_ips: %s",
        remote.my_ip, remote.dim_ips, remote.nm_ips
    )
    if options.num_islands == 1:
        REST_PORT = ISLAND_DEFAULT_REST_PORT

        # need to check for NM first and go on if co-hosted
        if remote.is_nm:
            co_hosted = remote.my_ip in remote.dim_ips
            _ = start_node_mgr(
                log_dir,
                remote.my_ip,
                logv=logv,
                max_threads=options.max_threads,
                host=None if options.all_nics else remote.my_ip,
                event_listeners=options.event_listeners,
                use_tool=co_hosted,
            )

        if remote.is_proxy:
            # Wait until the Island Manager is open
            if utils.portIsOpen(remote.hl_mgr_ip, ISLAND_DEFAULT_REST_PORT, 100):
                start_proxy(
                    remote.hl_mgr_ip,
                    ISLAND_DEFAULT_REST_PORT,
                    options.monitor_host,
                    options.monitor_port,
                )
            else:
                LOGGER.warning(
                    "Couldn't connect to the main drop manager, proxy not started"
                )
        elif remote.my_ip in remote.dim_ips:
            co_hosted = True
            nm_uris = [f"{ip}:{NODE_DEFAULT_REST_PORT}" for ip in remote.nm_ips]
            LOGGER.info("Starting island managers on nodes: %s", remote.dim_ips)
            _ = start_dim(nm_uris, log_dir, remote.my_ip, logv=logv)
            # whichever way we came from, now we have to wait until session is finished
            # we always monitor the island, else we will have race conditions
            physical_graph = get_pg(options, nm_uris, remote.dim_ips)
            monitoring_thread = submit_and_monitor(
                physical_graph,
                options,
                remote.dim_ips[0],
                REST_PORT,
                submit=co_hosted,
            )
            monitoring_thread.join()
            # now the session is finished

            # still shutting DIM down first to avoid monitoring conflicts
            stop_dims(remote.dim_ips)
            # now stop all the NMs
            stop_nms(remote.nm_ips)

        # shouldn't need this in addition
        # if dim_proc is not None:
        #     # Stop DALiuGE.
        #     LOGGER.info("Stopping DALiuGE island manager on rank %d", remote.rank)
        #     utils.terminate_or_kill(dim_proc, 5)

    elif remote.is_highest_level_manager:
        # TODO: In the case of more than one island the NMs are not yet started

        physical_graph = get_pg(options, remote.nm_ips, remote.dim_ips)
        remote.send_dim_nodes(physical_graph)

        # 7. make sure all DIMs are up running
        dim_ips_up = check_hosts(
            remote.dim_ips,
            timeout=MM_WAIT_TIME,
            retry=10,
        )
        if len(dim_ips_up) < len(remote.dim_ips):
            LOGGER.warning(
                "Not all DIMs were up and running: %d/%d",
                len(dim_ips_up),
                len(remote.dim_ips),
            )

        monitoring_thread = submit_and_monitor(
            physical_graph, options, remote.my_ip, MASTER_DEFAULT_REST_PORT
        )
        _ = start_mm(remote.dim_ips, log_dir, logv=logv)
        monitoring_thread.join()
        stop_mm(
            remote.my_ip
        )  # TODO: I don't think we need this and least not in the single island case
        stop_dims(remote.dim_ips)
    else:
        nm_ips = remote.recv_dim_nodes()
        proc = start_dim(nm_ips, log_dir, remote.my_ip, logv=logv)
        utils.wait_or_kill(proc, 1e8, period=5)
        stop_nms(remote.nm_ips)


if __name__ == "__main__":
    main()



# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '5.6.1'
git_version = '490762ca26f15c966a8b35a322ba91b27515c877'
full_version = '5.6.1'
is_release = True

if not is_release:
    version = full_version


#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2025
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import sys
import optparse # pylint: disable=deprecated-module
import shutil
import textwrap
import typing

from enum import Enum, auto
from dataclasses import dataclass
from importlib.resources import files
from pathlib import Path

import dlg.deploy.configs as default_configs
from dlg.deploy.configs import ConfigFactory

USER_CONFIG_DIR = Path.home() / ".config"
DLG_CONFIG_DIR = USER_CONFIG_DIR / "dlg"

class ConfigType(Enum):
    ENV = "ini"
    SLURM = "slurm"


class ConfigDirState(Enum):
    NEW = auto()
    EXISTS = auto()
    NOT_CREATED = auto()


@dataclass
class ConfigManager:
    """
    Utility class to manager job submission configuration files
    """

    facilities: typing.List[str]
    prompt_setup = True

    def setup_user(self):
        """
        Called when passing the "--setup" command of the "dlg remote" option.
        """
        dir_state = self.create_dlg_config_directory()
        if dir_state == ConfigDirState.NEW:
            self.copy_defaults_to_user_config_directory()
            self.print_available_config()
        elif dir_state == ConfigDirState.EXISTS:
            print(f"{DLG_CONFIG_DIR} already exists.")
        else:
            print(f"{DLG_CONFIG_DIR} was not created.")

    def prompt_user(self):
        """
        Create prompt and validate it against expected output
        """

        while True:
            ui = input("Do you want to create a $HOME/.config/dlg directory"
                       " to store your custom configuration files and scripts (y/n)? ")
            if ui.lower() == "y" or ui.lower() == "n":
                return ui.lower() == "y"
            else:
                print("Please selected from the options (Y/N)")

    def create_dlg_config_directory(self):
        """
        Establish the user $HOME/.config/dlg/ directory
        """
        if DLG_CONFIG_DIR.exists():
            return ConfigDirState.EXISTS
        else:
            if self.prompt_user():
                try:
                    Path.mkdir(DLG_CONFIG_DIR)
                    return ConfigDirState.NEW
                except Exception as e:
                    raise (e)
            else:
                return ConfigDirState.NOT_CREATED

    def copy_defaults_to_user_config_directory(self):
        """
        Move configuration files from dlg/deploy/configs to DLG_CONFIG_DIR
        """
        cfg_path = files(default_configs)

        configs = [p for p in cfg_path.iterdir() if
                   (".ini" in p.name) or (".slurm" in p.name)]

        if DLG_CONFIG_DIR.exists():
            for cfg in configs:
                if cfg.exists():
                    shutil.copy(cfg, DLG_CONFIG_DIR)
                else:
                    print("Unable to copy %s, does not exist", cfg)
        else:
            print("Unable to copy to %s, does is not available", DLG_CONFIG_DIR)

    def get_user_configs(self) -> typing.Dict:
        """
        Returns dictionary of filetypes to split out?
        """
        configs = {}
        if not DLG_CONFIG_DIR.exists():
            return configs
        filetypes = ["ini", "slurm"]

        for ftype in filetypes:
            configs[ftype] = [cfg for cfg in DLG_CONFIG_DIR.iterdir() if
                ftype in cfg.name]
        return configs

    def print_available_config(self):
        """
        Present the options available for deployment
        """
        print("User Configs (~/.config/dlg)")
        print("----------------------------")
        user_configs = self.get_user_configs()
        if user_configs:
            print("Environments (--config_file):")
            for config in user_configs["ini"]:
                print(textwrap.indent(config.name, "\t"))
            print("Slurm scripts (--slurm_template):")
            for config in user_configs["slurm"]:
                print(textwrap.indent(config.name, "\t"))
        else:
            print(textwrap.indent("N/a: User-specific directory is not setup.", "\t"))

        print("\nDALiuGE Defaults (-f/--facility):")
        print("-----------------------------------")
        for f in self.facilities:
            print(textwrap.indent(f, "\t"))

    def load_user_config(self, config_type: ConfigType, config_choice: str) -> Path:
        """
        Resolve the config type and choice to a config file and return full path.

        Path resolution occurs in the following hierarchy:
        - If config choice is absolute path, return the absolute path
        - If config choice is not an absolute path, but the path exists (i.e. can be
        found) return the path
        - If the config choice is relative, and doesn't exist, check if it exists in
          DLG_CONFIG_DIR
            - If it does exist, construct the path
            - If it doesn't exist, return None
        """
        if not DLG_CONFIG_DIR.exists() and self.prompt_setup:
            print("NOTE: No user configs exists; consider running 'dlg config --setup'.")
            self.prompt_setup = False

        choice_path = Path(config_choice)
        if choice_path.is_absolute() and choice_path.exists():
            return choice_path
        elif choice_path.exists():
            return choice_path.absolute()
        else:
            user_configs = self.get_user_configs()

            options = user_configs.get(config_type.value, [])
            for o in options:
                if config_choice == o.name:
                    return o
            return None

def run(_, args):

    cfg_manager = ConfigManager(ConfigFactory.available())

    parser = optparse.OptionParser()
    parser.add_option(
        "--setup",
        dest="setup",
        action="store_true",
        help="Setup local '$HOME/.config/dlg' directory to store custom environment config and slurm scripts",
        default=False
    )
    parser.add_option(
        "-l", "--list",
        dest="list",
        action="store_true",
        help="List the available configuration for DALiuGE deployment."
    )
    (opts, _) = parser.parse_args(sys.argv)
    if opts.setup:
        cfg_manager.setup_user()
        sys.exit(0)
    elif opts.list:
        print("Available facilities:\n")
        cfg_manager.print_available_config()
        sys.exit(0)
    else:
        print(f"Incorrect arguments: {args}")
        parser.print_help()






import logging
from dlg.manager import client

logger = logging.getLogger(f"dlg.{__name__}")


class NMAssigner:
    """
    Handles the logic of assigning / removing NMs from DIMs && handling new DIMs coming on/off-line.
    Maintains a leger of DIMs and NMs as name -> (server, port) mappings.
    Maintains a mapping of NMs to DIMs
    Allocation logic is currently very simple, handling only the case of a single DIM, but with
    room for more advanced load-balancing, if we require in the future.
    """

    def __init__(self):
        self.DIMs = {}  # Set of DIMs   name -> (server, port)
        self.NMs = {}  # Set of NMs     name -> (server, port)
        self.assignments = {}  # Maps NMs to DIMs
        self.mm_client = client.MasterManagerClient()

    def add_dim(self, name, server, port):
        self.DIMs[name] = (server, port)
        self.mm_client.add_dim(server)
        self.allocate_nms()

    def add_nm(self, name, server, port):
        self.NMs[name] = (server, port)
        self.mm_client.add_node(server)
        self.allocate_nms()

    def get_dim(self, name):
        return self.DIMs[name]

    def get_nm(self, name):
        return self.NMs[name]

    def remove_dim(self, name):
        server, port = self.DIMs[name]
        try:
            for nm in self.NMs:
                if self.assignments[nm] == server:
                    del self.assignments[nm]
            self.mm_client.remove_dim(server)
        finally:
            del self.DIMs[name]
        return server, port

    def remove_nm(self, name):
        server, _ = self.NMs[name]
        try:
            if name in self.assignments:
                dim_ip = self.assignments[name]
                nm_ip = self.NMs[name][0]
                self.mm_client.remove_node_from_dim(dim_ip, nm_ip)
            self.mm_client.remove_node(server)
        finally:
            del self.NMs[name]

    def allocate_nms(self):
        if self.DIMs == {}:
            logger.info("We have no DIMs")
        elif len(self.DIMs.keys()) == 1:
            dim_ip = list(self.DIMs.values())[0][0]  # IP of the first (only) DIM
            for nm, endpoint in self.NMs.items():
                nm_ip = endpoint[0]
                if nm not in self.assignments:
                    logger.info("Adding NM %s to DIM %s", nm_ip, dim_ip)
                    self.mm_client.add_node_to_dim(dim_ip, nm_ip)
                    self.assignments[nm] = dim_ip
                elif self.assignments[nm] not in self.DIMs:  # If we've removed a DIM
                    logger.info("Re-assigning %s to DIM %s", nm_ip, dim_ip)
                    self.mm_client.add_node_to_dim(dim_ip, nm_ip)
                    self.assignments[nm] = dim_ip
        else:  # We have lots of DIMs
            # Will do nothing, it's up to the user/deployer to handle this.
            logger.info("Multiple DIMs, handle node assignments externally.")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import json
import unittest
import daliuge_tests.engine.graphs as test_graphs

from pathlib import Path
from importlib.resources import files

from dlg.utils import getDlgWorkDir
from dlg.apps.app_base import BarrierAppDROP
from dlg.ddap_protocol import DROPLinkType, DROPStates, AppDROPStates
from dlg.droputils import DROPWaiterCtx
from dlg.exceptions import InvalidGraphException
from dlg.manager.session import SessionStates, Session

from test.dlg_engine_testutils import DROPManagerUtils


class TestSession(unittest.TestCase):
    def test_sessionStates(self):
        with Session("1") as s:
            self.assertEqual(SessionStates.PRISTINE, s.status)
            self.assertRaises(Exception, s.linkGraphParts, "", "", 0)
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata([{"oid": "A", "categoryType": "container"}])
            )
            self.assertEqual(SessionStates.BUILDING, s.status)

            s.deploy()
            self.assertEqual(SessionStates.RUNNING, s.status)

            # Now we can't do any of these
            self.assertRaises(Exception, s.deploy)
            self.assertRaises(Exception, s.addGraphSpec, "")
            self.assertRaises(Exception, s.linkGraphParts, "", "", 0)

    def test_sessionStates_noDrops(self):
        # No drops created, we can deploy right away
        with Session("1") as s:
            self.assertEqual(SessionStates.PRISTINE, s.status)
            s.deploy()
            self.assertEqual(SessionStates.FINISHED, s.status)

        with Session("2") as s:
            self.assertRaises(InvalidGraphException, s.deploy, completedDrops=["a"])

    def test_addGraphSpec(self):
        with Session("1") as s:
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata([{"oid": "A", "categoryType": "container"}])
            )
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata([{"oid": "B", "categoryType": "container"}])
            )
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata([{"oid": "C", "categoryType": "container"}])
            )

            # Adding an existing DROP
            self.assertRaises(
                Exception,
                s.addGraphSpec,
                DROPManagerUtils.add_test_reprodata([{"oid": "A", "categoryType": "container"}]),
            )

            # Adding invalid specs
            self.assertRaises(
                Exception,
                s.addGraphSpec,
                DROPManagerUtils.add_test_reprodata([{"oid": "D", "categoryType": "Application"}]),
            )  # missing "storage"
            self.assertRaises(
                Exception,
                s.addGraphSpec,
                DROPManagerUtils.add_test_reprodata(
                    [
                        {
                            "oid": "D",
                            "categoryType": "Data",
                            "dropclass": "invalid",
                        }
                    ]
                ),
            )  # invalid "storage"
            self.assertRaises(
                Exception,
                s.addGraphSpec,
                DROPManagerUtils.add_test_reprodata([{"oid": "D", "categoryType": "invalid"}]),
            )  # invalid "categoryType"
            # s.addGraphSpec(
            #     add_test_reprodata(
            #         [
            #             {
            #                 "oid": "D",
            #                 "categoryType": "Application",
            #                 "dropclass": "dlg.data.drops.NullDROP",
            #                 "outputs": ["X"],
            #             }
            #         ]
            #     ),
            # )
            # TODO: following test breaks if multi-node deployment works and vica versa
            #   There is clearly a conflict between detecting a real missing drop and a
            #   drop missing due to distribution.
            # self.assertRaises(
            #     Exception,
            #     s.addGraphSpec,
            #     add_test_reprodata(
            #         [
            #             {
            #                 "oid": "D",
            #                 "categoryType": "Application",
            #                 "dropclass": "dlg.data.drops.NullDROP",
            #                 "outputs": ["X"],
            #             }
            #         ]
            #     ),
            # )  # missing X DROP

    def test_addGraphSpec_namedPorts(self):
        with (files(test_graphs) / "funcTestPG_namedPorts.graph").open() as f:
           graphSpec = json.load(f)
        with Session("1") as s:
            s.addGraphSpec(graphSpec)
            s.deploy()

    def test_linking(self):
        with Session("1") as s:
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata([{"oid": "A", "categoryType": "container"}])
            )
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata(
                    [
                        {
                            "oid": "B",
                            "categoryType": "Application",
                            "dropclass": "dlg.apps.crc.CRCApp",
                        }
                    ]
                )
            )
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata([{"oid": "C", "categoryType": "container"}])
            )

            # Link them now
            s.linkGraphParts("A", "B", DROPLinkType.CONSUMER)
            s.linkGraphParts("B", "C", DROPLinkType.OUTPUT)

            # Deploy and check that the actual DROPs are linked together
            s.deploy()
            roots = s.roots
            self.assertEqual(1, len(roots))
            a = s.roots[0]
            self.assertEqual("A", a.oid)
            self.assertEqual(1, len(a.consumers))
            b = a.consumers[0]
            self.assertEqual("B", b.oid)
            self.assertEqual(1, len(b.outputs))
            c = b.outputs[0]
            self.assertEqual("C", c.oid)

    def test_cancel(self):
        """Cancels a whole graph execution"""
        with Session("1") as s:
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata(
                    [
                        {
                            "oid": "A",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "consumers": ["B"],
                        },
                        {
                            "oid": "B",
                            "categoryType": "Application",
                            "dropclass": "dlg.apps.simple.SleepApp",
                            "sleep_time": 2,
                        },
                        {
                            "oid": "C",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "producers": ["B"],
                        },
                    ]
                )
            )
            s.deploy()
            self.assertEqual(SessionStates.RUNNING, s.status)
            s.cancel()
            self.assertEqual(SessionStates.CANCELLED, s.status)
            for uid in "ABC":
                self.assertEqual(DROPStates.CANCELLED, s.drops[uid].status)
            self.assertEqual(AppDROPStates.CANCELLED, s.drops["B"].execStatus)

    def test_partial_cancel(self):
        """Like test_cancel, but only part of the graph should be cancelled"""
        with Session("1") as s:
            s.addGraphSpec(
                DROPManagerUtils.add_test_reprodata(
                    [
                        {
                            "oid": "A",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "consumers": ["B"],
                        },
                        {
                            "oid": "B",
                            "categoryType": "Application",
                            "dropclass": "dlg.apps.simple.SleepApp",
                            "sleep_time": 0,
                        },
                        {
                            "oid": "C",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "producers": ["B"],
                            "consumers": ["D"],
                        },
                        {
                            "oid": "D",
                            "categoryType": "Application",
                            "dropclass": "dlg.apps.simple.SleepApp",
                            "sleep_time": 10,
                        },
                        {
                            "oid": "E",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "producers": ["D"],
                        },
                    ]
                )
            )
            s.deploy()
            self.assertEqual(SessionStates.RUNNING, s.status)

            # Move first three drops to completed, D should take longer to execute
            with DROPWaiterCtx(self, s.drops["C"], 1):
                s.drops["A"].write(b"x")
                s.drops["A"].setCompleted()
            # Cancel the session, A, B and C should remain COMPLETED
            s.cancel()
            self.assertEqual(SessionStates.CANCELLED, s.status)
            for uid in "ABC":
                self.assertEqual(DROPStates.COMPLETED, s.drops[uid].status)
            for uid in "DE":
                self.assertEqual(DROPStates.CANCELLED, s.drops[uid].status)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
A module containing utility code for running remote commands over SSH.
"""

import logging
import sys
import time

from paramiko.client import SSHClient, AutoAddPolicy
import scp

logger = logging.getLogger(f"dlg.{__name__}")


def execRemoteWithClient(client, command, timeout=None, bufsize=-1):
    """
    Executes `command` using the given SSH `client`.
    """
    chan = client.get_transport().open_session()
    chan.settimeout(timeout)
    chan.exec_command('/bin/bash -l -c "%s"' % (command.replace('"', '"')))
    # Otherwise do something like this:
    # chan.get_pty(width=80, height=24)
    # chan.invoke_shell()
    # chan.sendall(command + "\n")

    # Wait until the command has finished execution and return the full contents
    # of the stdout and stderr
    while True:
        if chan.exit_status_ready():
            break
        try:
            time.sleep(0.01)
        except KeyboardInterrupt:
            chan.send("\x03")

    exitStatus = chan.recv_exit_status()
    with chan.makefile("r", bufsize) as f:
        stdout = f.read()
    with chan.makefile_stderr("r", bufsize) as f:
        stderr = f.read()
    return stdout, stderr, exitStatus


def execRemote(
    host: str,
    command: str,
    username: str = None,
    pkeyPath: str = None,
    timeout: int = None,
    bufsize: int = -1,
):
    """
    Executes `command` on `host`. If `username` is provided, the command will
    be run as `username`; otherwise the current user will be used to run the
    command. Returns the tuple (stdout, stderr, exitStatus)

    Parameters:
    -----------
    host: str, Name of the host to connect to
    command: str, Command to execute
    username: str, Username to be used for connection, if None current user is used
    pkeyPath: str, Path to private key file for connection
    timeout: int, Timeout in seconds, None is default
    bufsize: int, Size of buffer for the return, unlimited by default

    Returns:
    --------
    tuple, (stdout, stderr, exitStatus)
    """
    with createClient(host, username, pkeyPath=pkeyPath) as client:
        return execRemoteWithClient(client, command, timeout, bufsize)


def createClient(host, username=None, pkeyPath=None):
    """
    Creates an SSH client object that can be used to perform SSH-related
    operations
    """
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy())
    try:
        client.connect(host, username=username, key_filename=pkeyPath)
    except FileNotFoundError:
        logger.warning(
            "File '%s' was not found, cannot create remote connection", pkeyPath
        )
        sys.exit(1)
    return client


def __scpProgress(filename, size, sent):
    if size == sent:
        logger.debug("Finished scp-ing %s (%d bytes)", filename, sent)


def copyFrom(
    host,
    remotePath,
    localPath=".",
    recursive=False,
    username=None,
    pkeyPath=None,
    timeout=None,
):
    """
    Copies the files located at `host`:`remotePath` to `localPath` connecting
    to `host` as `username`. A `recursive` flag can be specified, as well as a
    private key (via `pkeyPath`) to be used when creating the connection.
    """
    if timeout is None:
        timeout = 5.0
    client = createClient(host, username=username, pkeyPath=pkeyPath)
    with scp.SCPClient(
        client.get_transport(), progress=__scpProgress, socket_timeout=timeout
    ) as scpClient:
        scpClient.get(remote_path=remotePath, local_path=localPath, recursive=recursive)
    client.close()


def copyTo(
    host,
    localFiles,
    remotePath=".",
    recursive=False,
    username=None,
    pkeyPath=None,
    timeout=None,
):
    """
    Copies the files located at `localPath` to `host`:`remotePath` connecting
    to `host` as `username`. A `recursive` flag can be specified, as well as a
    private key (via `pkeyPath`) to be used when creating the connection.
    """
    if timeout is None:
        timeout = 5.0
    client = createClient(host, username=username, pkeyPath=pkeyPath)
    with scp.SCPClient(
        client.get_transport(), progress=__scpProgress, socket_timeout=timeout
    ) as scpClient:
        scpClient.put(localFiles, remote_path=remotePath, recursive=recursive)
    client.close()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import errno
import os

from dlg.common.reproducibility.reproducibility import common_hash
from dlg.data import path_builder
from dlg.data.io import FileIO
from dlg.ddap_protocol import DROPStates
from .data_base import DataDROP, PathBasedDrop, logger, track_current_drop
from dlg.exceptions import InvalidDropException
from dlg.meta import dlg_bool_param
from dlg.utils import isabs
from typing import Union


##
# @brief File
# @details A standard file on a filesystem mounted to the deployment machine
# @par EAGLE_START
# @param category File
# @param tag daliuge
# @param filepath /String/ApplicationArgument/NoPort/ReadWrite//False/False/"File path for this file. In many cases this does not need to be specified. If it has a \/ at the end it will be treated as a directory name and the filename will be generated. If it does not have a \/, the last part will be treated as a filename. If filepath does not start with \/ (relative path) then the session directory will be prepended to make the path absolute.""
# @param check_filepath_exists False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Perform a check to make sure the file path exists before proceeding with the application
# @param dropclass dlg.data.drops.file.FileDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name file/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param persist True/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class FileDROP(DataDROP, PathBasedDrop):
    """
    A DROP that points to data stored in a mounted filesystem.

    Users can fix both the path and the name of a FileDrop using the `filepath`
    parameter for each FileDrop. We distinguish four cases and their combinations.

    1) If not specified the filename will be generated.
    2) If it has a '/' at the end it will be treated as a directory name and the
       filename will the generated.
    3) If it does not end with a '/' and it is not an existing directory, it is
       treated as dirname plus filename.
    4) If filepath points to an existing directory, the filename will be generated

    In all cases above, if `filepath` does not start with '/’ (relative path)
    then the session directory will be pre-pended to make the path absolute.
    """

    delete_parent_directory = dlg_bool_param("delete_parent_directory", False)
    check_filepath_exists = dlg_bool_param("check_filepath_exists", False)
    # is_dir = dlg_bool_param("is_dir", False)

    # Make sure files are not deleted by default and certainly not if they are
    # marked to be persisted no matter what expireAfterUse said
    def __init__(self, *args, **kwargs):
        if "persist" not in kwargs:
            kwargs["persist"] = True
        if kwargs["persist"] and "lifespan" not in kwargs:
            kwargs["expireAfterUse"] = False
        self.is_dir = False
        self._updatedPorts = False
        super().__init__(*args, **kwargs)

    def sanitize_paths(self, filepath: str) -> Union[None, str]:
        """
        Expand ENV_VARS, but also deal with relative
        and absolute paths. filepath can be either just be
        a directory, a directory including a file name, only
        a directory (both relative and absolute), or just
        a file name.

        :param filepath: string, path and or directory

        :returns filepath
        """
        # replace any ENV_VARS on the names
        if not filepath:
            return None
        filepath = os.path.expandvars(filepath)
        if isabs(filepath):
            return filepath
        else:
            filepath = self.get_dir(filepath)
        return filepath


    def initialize(self, **kwargs):
        """
        FileDROP-specific initialization.
        """

        self._setupFilePaths()

    def _setupFilePaths(self):
        filepath = self.parameters.get("filepath", None)
        # TODO ADD SUFFIX/PREFIX
        dirname = None
        filename = None

        if filepath:  # if there is anything provided
            # TODO do f-string substitution if necessary
            if "/" not in filepath:  # just a name
                filename = filepath
                dirname = self.get_dir(".")
            # filepath = self.sanitize_paths(self.filepath)
            elif filepath.endswith("/"):  # just a directory name
                self.is_dir = True
                filename = None
                dirname = filepath
            else:
                filename = os.path.basename(filepath)
                dirname = os.path.dirname(filepath)
        if dirname is None:
            dirname = "."
        filename = os.path.expandvars(filename) if filename else None
        dirname = self.sanitize_paths(dirname) if dirname else None
        # We later check if the file exists, but only if the user has specified
        # an absolute dirname/filepath (otherwise it doesn't make sense, since
        # we create our own filenames/dirnames dynamically as necessary
        check = False
        if isabs(dirname):
            check = self.check_filepath_exists

        # Default filename to drop human readable format based on UID
        if filename is None:
            filename = path_builder.base_uid_filename(self.uid, self._humanKey)


        self.filename = filename
        self.dirname = self.get_dir(dirname)
        self._root = self.dirname
        self._path = (
            os.path.join(self.dirname, self.filename) if self.filename else self.dirname
        )
        logger.debug(
            "Set path of drop %s: %s check: %s %s",
            self._uid, self._path, check, os.path.isfile(self._path)
        )
        if check and not os.path.isfile(self._path):
            raise InvalidDropException(
                self, "File does not exist or is not a file: %s" % self._path
            )

        self._wio = None

    def getIO(self):

        # We need to update named_ports now we have runtime information
        # if not self._updatedPorts:
        if not self.parameters.get("filepath", None):
            self._map_input_ports_to_params()
            self._setupFilePaths()

        return FileIO(self._path)

    def delete(self):
        super().delete()
        if self.delete_parent_directory:
            try:
                os.rmdir(self._root)
            except OSError as e:
                # Silently ignore "Directory not empty" errors
                if e.errno != errno.ENOTEMPTY:
                    raise

    @track_current_drop
    def setCompleted(self):
        """
        Override this method in order to get the size of the drop set once it is completed.
        """
        # TODO: This implementation is almost a verbatim copy of the base class'
        # so we should look into merging them
        status = self.status
        if status == DROPStates.CANCELLED:
            return
        elif status == DROPStates.SKIPPED:
            self._fire("dropCompleted", status=status)
            return
        elif status not in [
            DROPStates.COMPLETED,
            DROPStates.INITIALIZED,
            DROPStates.WRITING,
        ]:
            raise Exception(
                "%r not in INITIALIZED or WRITING state (%s), cannot setComplete()"
                % (self, self.status)
            )

        self._closeWriters()

        if status != DROPStates.COMPLETED:
            logger.debug("Moving %r to COMPLETED", self)
            self.status = DROPStates.COMPLETED

        # here we set the size. It could happen that nothing is written into
        # this file, in which case we create an empty file so applications
        # downstream don't fail to read
        try:
            self._size = os.stat(self.path).st_size
        except FileNotFoundError:
            # we''ll try this again in case there is some other issue
            try:
                with open(self.path, "wb"):
                    pass
            except IOError:
                self.status = DROPStates.ERROR
                logger.error("Path not accessible: %s", self.path)
            self._size = 0
        # Signal our subscribers that the show is over
        self._fire("dropCompleted", status=DROPStates.COMPLETED)
        self.completedrop()

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]  # TODO: change when necessary
        return "file://" + hostname + self._path

    # Override
    def generate_reproduce_data(self):
        from dlg.droputils import allDropContents

        try:
            data = allDropContents(self, self.size)
        except IOError:
            data = b""
        return {"data_hash": common_hash(data)}



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
RPC support for DALiuGE

This module contains all client and server RPC classes for all the different
technologies we support.
"""

import collections
import dataclasses
import logging
import queue
import threading

import gevent
import zerorpc

from dlg.manager.manager_data import Node

from dlg import utils

logger = logging.getLogger(f"dlg.{__name__}")


class RPCObject(object):
    """Base class for all RCP clients and server"""

    def start(self):
        self.rpc_running = True

    def shutdown(self):
        self.rpc_running = False


class RPCClientBase(RPCObject):
    """Base class for all RPC clients"""

    def get_drop_attribute(self, hostname, port, session_id, uid, name):

        # hostname = hostname.split(":")[0]

        logger.debug(
            "Getting attribute %s for drop %s of session %s at %s:%d",
            name,
            uid,
            session_id,
            hostname,
            port,
        )
        # hostname = hostname.split(":")[0]

        client, closer = self.get_rpc_client(hostname, port)

        # The remote method receives the same client used to inspect the remote
        # object, and it closes it when the method is not used anymore
        class remote_method(object):
            def __del__(self):
                closer()

            def __call__(self, *args):
                return client.call_drop(session_id, uid, name, *args)

        # Shortcut to avoid extra calls
        known_methods = ()
        # known_methods = ('open', 'read', 'write', 'close')
        closeit = False
        try:
            if name in known_methods or client.has_method(session_id, uid, name):
                return remote_method()
            closeit = True
            return client.get_drop_property(session_id, uid, name)
        finally:
            if closeit:
                closer()


class RPCServerBase(RPCObject):
    """Base class for all RPC server"""

    def __init__(self, host, port):
        self._rpc_host = host.split(":")[0]
        self._rpc_port = port


class ZeroRPCClient(RPCClientBase):
    """ZeroRPC client support"""

    request = collections.namedtuple("request", "method args queue")
    response = collections.namedtuple("response", "value is_exception")

    def __init__(self, *args, **kwargs):
        super(ZeroRPCClient, self).__init__(*args, **kwargs)
        self._zrpcclients = {}
        self._zrpcclientthreads = []
        self._own_context = False
        logger.debug("RPC Client created")

    def __del__(self):
        if self._own_context and self._context:
            self._context.term()

    def start(self):
        super(ZeroRPCClient, self).start()
        if not hasattr(self, "_context"):
            self._context = zerorpc.Context()
            self._own_context = True

        # One per remote host
        self._zrpcclient_acquisition_lock = threading.Lock()
        self._zrpcclients = {}
        self._zrpcclientthreads = []

    def shutdown(self):
        super(ZeroRPCClient, self).shutdown()
        for t in self._zrpcclientthreads:
            t.join(10)
            if t.is_alive():
                logger.warning("ZeroRPC client thread %s is still alive", t.name)
        if self._own_context:
            self._context.term()
            self._context = None

    def get_client_for_endpoint(self, host, port):

        # host = host.split(":")[0]
        if isinstance(host, Node):
            endpoint = (host.host, port)
        else:
            endpoint = (host, port)

        with self._zrpcclient_acquisition_lock:
            if endpoint in self._zrpcclients:
                return self._zrpcclients[endpoint]

            # We start the new client on its own thread so it uses gevent, etc.
            # In this thread we create simply enqueue requests
            req_queue = queue.Queue()
            tname_tpl, args = "zrpc(%s:%d)", (host, port)
            t = threading.Thread(
                target=self.run_zrpcclient,
                args=(host, port, req_queue),
                name=tname_tpl % args,
            )
            t.start()

            class QueueingClient(object):
                def __make_call(self, method, *args):
                    res_queue = queue.Queue()
                    request = ZeroRPCClient.request(method, args, res_queue)
                    req_queue.put(request)
                    x = res_queue.get()
                    if x.is_exception:
                        raise x.value
                    return x.value

                def call_drop(self, session_id, uid, name, *args):
                    return self.__make_call("call_drop", session_id, uid, name, *args)

                def get_drop_property(self, session_id, uid, name):
                    return self.__make_call("get_drop_property", session_id, uid, name)

                def has_method(self, session_id, uid, name):
                    return self.__make_call("has_method", session_id, uid, name)

            client = QueueingClient()
            self._zrpcclients[endpoint] = client
            self._zrpcclientthreads.append(t)
            return client

    def run_zrpcclient(self, host, port, req_queue):
        if isinstance(host, Node):
            host = host.host  # split(":")[0]
        client = zerorpc.Client("tcp://%s:%d" % (host, port), context=self._context)

        forwarder = gevent.spawn(self.forward_requests, req_queue, client)
        gevent.joinall([forwarder])

        logger.info("Closing %s:%d ZeroRPC client", host, port)
        client.close()

    def forward_requests(self, req_queue, client):
        while self.rpc_running:
            try:
                req = req_queue.get_nowait()
                gevent.spawn(self.queue_request, client, req)
            except queue.Empty:
                gevent.sleep(0.005)

    def process_response(self, req, async_response):
        try:
            x = ZeroRPCClient.response(async_response.get_nowait(), False)
        except gevent.Timeout as e:
            raise RuntimeError("Timed out on AsyncResult.get_nowait") from e
        except RuntimeError as e:
            x = ZeroRPCClient.response(e, True)
        req.queue.put(x)

    def queue_request(self, client, req):
        # Pass "async" in a dictionary; 3.7+ fails because it's a keyword
        async_result = client.__call__(req.method, *req.args, **{"async": True})
        async_result.rawlink(lambda x: self.process_response(req, x))

    def get_rpc_client(self, hostname, port):
        # hostname = hostname.split(":")[0]
        client = self.get_client_for_endpoint(hostname, port)
        # No closing function since clients are long-lived
        return client, lambda: None


class ZeroRPCServer(RPCServerBase):
    """ZeroRPC server support"""

    @classmethod
    def create_context(cls):
        # This import can take a long time in big HPC deployments
        return zerorpc.Context()

    @property
    def rpc_endpoint(self):
        return self._rpc_host, self._rpc_port

    def start(self):
        super(ZeroRPCServer, self).start()

        # Starts the single-threaded ZeroRPC server for RPC requests
        timeout = 30
        server_started = threading.Event()
        self._zrpcserverthread = threading.Thread(
            target=self.run_zrpcserver,
            name="ZeroRPC server",
            args=(self._rpc_host, self._rpc_port, server_started),
        )
        self._zrpcserverthread.start()
        if not server_started.wait(timeout):
            raise Exception(
                "ZeroRPC server didn't start within %d seconds" % (timeout,)
            )

    def run_zrpcserver(self, host, port, server_started):
        # Use context provided by subclass
        self._zrpcserver = zerorpc.Server(self, context=self._context)
        # zmq needs an address, not a hostname
        endpoint = "tcp://%s:%d" % (utils.zmq_safe(host), port)
        logger.debug("Trying to bind ZeroRPC to %s", endpoint)
        self._zrpcserver.bind(endpoint)
        logger.info("Listening for RPC requests via ZeroRPC on %s", endpoint)
        server_started.set()

        runner = gevent.spawn(self._zrpcserver.run)
        stopper = gevent.spawn(self.stop_zrpcserver)
        gevent.joinall([runner, stopper])
        logger.info("ZeroRPC server finished")

    def stop_zrpcserver(self):
        while self.rpc_running:
            gevent.sleep(0.01)
        logger.info(
            "Closing ZeroRPC server on tcp://%s:%d",
            utils.zmq_safe(self._rpc_host),
            self._rpc_port,
        )
        self._zrpcserver.close()

    def shutdown(self):
        super(ZeroRPCServer, self).shutdown()
        self._zrpcserverthread.join()


RPCServer, RPCClient = ZeroRPCServer, ZeroRPCClient


@dataclasses.dataclass(frozen=True)
class ProxyInfo:
    """Information needed to create a DropProxy"""

    hostname: str
    port: int
    session_id: str
    uid: str

    @classmethod
    def from_data_drop(cls, drop):
        if isinstance(drop, DropProxy):
            return drop.proxy_info

        # TODO: we can't use the RPC endpoint's host directly here, as that
        #       indicates the address the different servers *bind* to
        #       (and, for example, can be 0.0.0.0)
        assert drop.rpc_endpoint
        rpc_host, rpc_port = drop.rpc_endpoint
        rpc_host = utils.to_externally_contactable_host(rpc_host, prefer_local=True)
        return cls(rpc_host, rpc_port, drop.dlg_session_id, drop.uid)

    def __repr__(self):
        return (
            f"<ProxyInfo {self.uid}, session {self.session_id} "
            f"@{self.hostname}:{self.port}>"
        )


class DropProxy(object):
    """
    A proxy to a remote drop.

    It forwards attribute requests and procedure calls through the given RPC client.
    """

    def __init__(self, rpc_client, proxy_info: ProxyInfo):
        self.rpc_client = rpc_client
        self._proxy_info: ProxyInfo = proxy_info
        logger.debug("Created %r", self)

    def handleEvent(self, evt):
        pass

    def __getattr__(self, name):
        if name == "uid":
            return self._proxy_info.uid
        elif name in ("inputs", "streamingInputs", "outputs", "consumers", "producers"):
            return []
        return self.rpc_client.get_drop_attribute(
            self._proxy_info.hostname,
            self._proxy_info.port,
            self._proxy_info.session_id,
            self.uid,
            name,
        )

    def __repr__(self):
        return f"<DropProxy with {self._proxy_info}"

    @property
    def proxy_info(self) -> ProxyInfo:
        """
        Get ProxyInfo
        :return:  the proxy info associated with the object
        """
        return self._proxy_info



# pylint: skip-file
"""
Copyright (C) 2015 Ryan Gonzalez

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


g_backup = globals().copy()

__version__ = '0.8'

__all__ = ['overload', 'RuntimeModule', 'switch', 'tail_recurse', 'copyfunc',
           'set_docstring', 'annotate', 'safe_unpack', 'modify_function',
           'assign',  'fannotate', 'compare_and_swap', 'is_main',
           'call_if_main', 'run_main']

import sys, inspect, types, functools

# Credit @clefourrier
# https://github.com/refi64/PyExt/pull/3/
if not hasattr(inspect, 'getargspec'):
    inspect.getargspec = inspect.getfullargspec

def _targspec(func, specs, attr='__orig_arg__'):
    if hasattr(func, '__is_overload__') and func.__is_overload__:
        return getattr(func, attr)
    return specs(func)

def set_docstring(doc):
    '''A simple decorator to set docstrings.

       :param doc: The docstring to tie to the function.

       Example::

          @set_docstring('This is a docstring')
          def myfunc(x):
              pass'''
    def wrap(f):
        f.__doc__ = doc
        return f
    return wrap

_modify_function_doc = '''
Creates a copy of a function, changing its attributes.

:param globals: Will be added to the function's globals.

:param name: The new function name. Set to ``None`` to use the function's original name.

:param code: The new function code object. Set to ``None`` to use the function's original code object.

:param defaults: The new function defaults. Set to ``None`` to use the function's original defaults.

:param closure: The new function closure. Set to ``None`` to use the function's original closure.

.. warning:: This function can be potentially dangerous.
'''

def copyfunc(f):
   '''Copies a funcion.

      :param f: The function to copy.

      :return: The copied function.

      .. deprecated:: 0.4
         Use :func:`modify_function` instead.
      '''
   return modify_function(f)

if sys.version_info.major == 3:
    @set_docstring(_modify_function_doc)
    def modify_function(f, globals={}, name=None, code=None, defaults=None,
                        closure=None):
        if code is None: code = f.__code__
        if name is None: name = f.__name__
        if defaults is None: defaults = f.__defaults__
        if closure is None: closure = f.__closure__
        newf = types.FunctionType(code, dict(f.__globals__, **globals), name=name,
                                  argdefs=defaults, closure=closure)
        newf.__dict__.update(f.__dict__)
        return newf
    argspec = inspect.getfullargspec
    ofullargspec = inspect.getfullargspec
    def _fullargspec(func):
        return _targspec(func, ofullargspec)
    inspect.getfullargspec = _fullargspec
    def _exec(m,g): exec(m,g)
else:
    @set_docstring(_modify_function_doc)
    def modify_function(f, globals={}, name=None, code=None, defaults=None,
                        closure=None):
        if code is None: code = f.func_code
        if name is None: name = f.__name__
        if defaults is None: defaults = f.func_defaults
        if closure is None: closure = f.func_closure
        newf = types.FunctionType(code, dict(f.func_globals, **globals),
                                  name=name, argdefs=defaults, closure=closure)
        newf.__dict__.update(f.__dict__)
        return newf
    argspec = inspect.getargspec
    eval(compile('def _exec(m,g): exec m in g', '<exec>', 'exec'))

def _gettypes(args):
    return tuple(map(type, args))

oargspec = inspect.getargspec

def _argspec(func):
    return _targspec(func, oargspec)

inspect.getargspec = _argspec


class overload(object):
    '''Simple function overloading in Python.'''
    @classmethod
    def argc(self, argc=None):
        '''Overloads a function based on the specified argument count.

           :param argc: The argument count. Defaults to ``None``. If ``None`` is given, automatically compute the argument count from the given function.

           .. note::

              Keyword argument counts are NOT checked! In addition, when the argument count is automatically calculated, the keyword argument count is also ignored!

           Example::

               @overload.argc()
               def func(a):
                   print 'Function 1 called'

               @overload.argc()
               def func(a, b):
                   print 'Function 2 called'

               func(1) # Calls first function
               func(1, 2) # Calls second function
               func() # Raises error
               '''
        # Python 2 UnboundLocalError fix
        argc = {'argc': argc}
        def wrap(f):
            if argc['argc'] is None:
                argc['argc'] = len(argspec(f).args)
            try:
                st = inspect.stack()[1][0]
                oldf = dict(st.f_globals, **st.f_locals)[f.__name__]
            except KeyError: pass
            else:
                if hasattr(oldf, '__pyext_overload_basic__'):
                    globls = oldf.__globals__ if sys.version_info.major == 3\
                             else oldf.func_globals
                    globls['overloads'][argc['argc']] = f
                    return oldf
            @functools.wraps(f)
            def newf(*args, **kwargs):
                if len(args) not in overloads:
                    raise TypeError(
                        "No overload of function '%s' that takes %d args" % (
                            f.__name__, len(args)))
                return overloads[len(args)](*args, **kwargs)
            overloads = {}
            overloads[argc['argc']] = f
            newf = modify_function(newf, globals={'overloads': overloads})
            newf.__pyext_overload_basic__ = None
            newf.__orig_arg__ = argspec(f)
            return newf
        return wrap
    @classmethod
    def args(self, *argtypes, **kw):
        '''Overload a function based on the specified argument types.

           :param argtypes: The argument types. If None is given, get the argument types from the function annotations(Python 3 only)
           :param kw: Can only contain 1 argument, `is_cls`. If True, the function is assumed to be part of a class.

           Example::

               @overload.args(str)
               def func(s):
                   print 'Got string'

               @overload.args(int, str)
               def func(i, s):
                   print 'Got int and string'

               @overload.args()
               def func(i:int): # A function annotation example
                   print 'Got int'

               func('s')
               func(1)
               func(1, 's')
               func(True) # Raises error
            '''

        # XXX: some of this should be moved to a utility class
        # It's duplicated from overload.argc
        # Python 2 UnboundLocalError fix...again!
        argtypes = {'args': tuple(argtypes)}
        def wrap(f):
            if len(argtypes['args']) == 1 and argtypes['args'][0] is None:
                aspec = argspec(f)
                argtypes['args'] = tuple(map(lambda x: x[1], sorted(
                    aspec.annotations.items(),
                    key=lambda x: aspec.args.index(x[0]))))
            try:
                st = inspect.stack()[1][0]
                oldf = dict(st.f_globals, **st.f_locals)[f.__name__]
            except KeyError: pass
            else:
                if hasattr(oldf, '__pyext_overload_args__'):
                    globls = oldf.__globals__ if sys.version_info.major == 3\
                             else oldf.func_globals
                    globls['overloads'][argtypes['args']] = f
                    return oldf
            @functools.wraps(f)
            def newf(*args):
                if len(kw) == 0:
                    cargs = args
                elif len(kw) == 1 and 'is_cls' in kw and kw['is_cls']:
                    cargs = args[1:]
                else:
                    raise ValueError('Invalid keyword args specified')
                types = _gettypes(cargs)
                if types not in overloads:
                    raise TypeError(\
                        "No overload of function '%s' that takes: %s" % (
                            f.__name__, types))
                return overloads[types](*args)
            overloads = {}
            overloads[argtypes['args']] = f
            newf = modify_function(newf, globals={'overloads': overloads})
            newf.__pyext_overload_args__ = None
            newf.__orig_arg__ = argspec(f)
            return newf
        return wrap

class _RuntimeModule(object):
    'Create a module object at runtime and insert it into sys.path. If called, same as :py:func:`from_objects`.'
    def __call__(self, *args, **kwargs):
        return self.from_objects(*args, **kwargs)
    @staticmethod
    @overload.argc(1)
    def from_objects(name, **d):
        return _RuntimeModule.from_objects(name, '', **d)
    @staticmethod
    @overload.argc(2)
    def from_objects(name, docstring, **d):
        """Create a module at runtime from `d`.

           :param name: The module name.

           :param docstring: Optional. The module's docstring.

           :param d: All the keyword args, mapped from name->value.

           Example: ``RuntimeModule.from_objects('name', 'doc', a=1, b=2)``"""
        module = types.ModuleType(name, docstring)
        module.__dict__.update(d)
        module.__file__ = '<runtime_module>'
        sys.modules[name] = module
        return module
    @staticmethod
    @overload.argc(2)
    def from_string(name, s):
        return _RuntimeModule.from_string(name, '', s)
    @staticmethod
    @overload.argc(3)
    def from_string(name, docstring, s):
        '''Create a module at runtime from `s``.

           :param name: The module name.

           :param docstring: Optional. The module docstring.

           :param s: A string containing the module definition.'''
        g = {}
        _exec(s, g)
        return _RuntimeModule.from_objects(name, docstring,
                    **dict(filter(lambda x: x[0] not in g_backup, g.items())))

RuntimeModule = _RuntimeModule()

class CaseObject(object):
    'The object returned by a switch statement. When called, it will return True if the given argument equals its value, else False. It can be called with multiple parameters, in which case it checks if its value equals any of the arguments.'
    def __init__(self, value, cstyle):
        self.value = value
        self.did_match = False
        self.cstyle = cstyle
        self.did_pass = not cstyle
    def __call__(self, *args):
        if not self.cstyle and self.did_match: return False
        if assign('res', not (self.did_pass and self.cstyle) and\
                         self.value in args):
            self.did_match = True
        return res
    def quit(self):
        'Forces all other calls to return False. Equilavent of a ``break`` statement.'
        self.did_pass = True
    def default(self):
        "Executed if ``quit`` wasn't called."
        return not self.did_match and (not self.did_pass if self.cstyle else True)
    def __iter__(self):
        yield self
    def __enter__(self):
        return self
    def __exit__(self, *args):
        pass

def switch(value, cstyle=False):
    '''A Python switch statement implementation that is used with a ``with`` statement.

       :param value: The value to "switch".
       :param cstyle: If ``True``, then cases will automatically fall through to the next one until ``case.quit()`` is encountered.

       ``with`` statement example::

           with switch('x'):
               if case(1): print 'Huh?'
               if case('x'): print 'It works!!!'

       .. warning:: If you modify a variable named "case" in the same scope that you use the ``with`` statement version, you will get an UnboundLocalError. The soluction is to use ``with switch('x') as case:`` instead of ``with switch('x'):``.'''
    res = CaseObject(value, cstyle)
    inspect.stack()[1][0].f_globals['case'] = res
    return res

def tail_recurse(spec=None):
    '''Remove tail recursion from a function.

       :param spec: A function that, when given the arguments, returns a bool indicating whether or not to exit. If ``None,`` tail recursion is always called unless the function returns a value.

       .. note::

           This function has a slight overhead that is noticable when using timeit. Only use it if the function has a possibility of going over the recursion limit.

       .. warning::

           This function will BREAK any code that either uses any recursion other than tail recursion or calls itself multiple times. For example, ``def x(): return x()+1`` will fail.

       Example::

           @tail_recurse()
           def add(a, b):
               if a == 0: return b
               return add(a-1, b+1)

           add(10000000, 1) # Doesn't max the recursion limit.
           '''
    def _wrap(f):
        class TailRecursion(Exception):
            def __init__(self, args, kwargs):
                self.args = args
                self.kwargs = kwargs
        def _newf(*args, **kwargs):
            if inspect.stack()[1][3] == f.__name__:
                if (spec and spec(args)) or not spec:
                    raise TailRecursion(args, kwargs)
            while True:
                try:
                    res = f(*args, **kwargs)
                except TailRecursion as ex:
                    args = ex.args
                    kwargs = ex.kwargs
                    continue
                else:
                    return res
        _newf.__doc__ = f.__doc__
        return _newf
    return _wrap

def annotate(*args, **kwargs):
    '''Set function annotations using decorators.

       :param args: This is a list of annotations for the function, in the order of the function's parameters. For example, ``annotate('Annotation 1', 'Annotation 2')`` will set the annotations of parameter 1 of the function to ``Annotation 1``.

       :param kwargs: This is a mapping of argument names to annotations. Note that these are applied *after* the argument list, so any args set that way will be overriden by this mapping. If there is a key named `ret`, that will be the annotation for the function's return value.

       .. deprecated:: 0.5
         Use :func:`fannotate` instead.
'''
    def _wrap(f):
        if not hasattr(f, '__annotations__'):
            f.__annotations__ = {}
        if 'ret' in kwargs:
            f.__annotations__['return'] = kwargs.pop('ret')
        f.__annotations__.update(dict(zip(argspec(f).args, args)))
        f.__annotations__.update(kwargs)
        return f
    return _wrap

def fannotate(*args, **kwargs):
    """Set function annotations using decorators.

       :param args: The first positional argument is used for the function's return value; all others are discarded.

       :param kwargs: This is a mapping of argument names to annotations.

       Example::

           @fannotate('This for the return value', a='Parameter a', b='Parameter b')
           def x(a, b):
               pass

       """
    def wrap(f):
        if not hasattr(f, '__annotations__'):
            f.__annotations__ = {}
        if len(args) >= 1:
            f.__annotations__['return'] = args[0]
        f.__annotations__.update(kwargs)
        return f
    return wrap

def safe_unpack(seq, ln, fill=None):
    '''Safely unpack a sequence to length `ln`, without raising ValueError. Based on Lua's method of unpacking. Empty values will be filled in with `fill`, while any extra values will be cut off.

       :param seq: The sequence to unpack.

       :param ln: The expected length of the sequence.

       :param fill: The value to substitute if the sequence is too small. Defaults to ``None``.

       Example::

           s = 'a:b'
           a, b = safe_unpack(s.split(':'), 2)
           # a = 'a'
           # b = 'b'
           s = 'a'
           a, b = safe_unpack(s.split(':'), 2)
           # a = 'a'
           # b = None'''
    if len(seq) > ln:
        return seq[:ln]
    elif len(seq) < ln:
        return seq + type(seq)([fill]*(ln-len(seq)))
    else:
        return seq

def assign(varname, value):
    '''Assign `value` to `varname` and return it. If `varname` is an attribute and the instance name it belongs to is not defined, a NameError is raised.
       This can be used to emulate assignment as an expression. For example, this::

          if assign('x', 7): ...

       is equilavent to this C code::

          if (x = 7) ...

       .. warning::

          When assigning an attribute, the instance it belongs to MUST be declared as global prior to the assignment. Otherwise, the assignment will not work.
    '''
    fd = inspect.stack()[1][0].f_globals
    fl = inspect.stack()[1][0].f_locals
    if '.' not in varname:
        fd[varname] = value
    else:
        vsplit = list(map(str.strip, varname.split('.')))
        fvars = dict(fd, **fl)
        if vsplit[0] not in fvars:
            raise NameError('Unknown object: %s' % vsplit[0])
        base = fvars[vsplit[0]]
        for x in vsplit[1:-1]:
            base = getattr(base, x)
        setattr(base, vsplit[-1], value)
    return value

def is_main(frame=1):
    "Return if the caller is main. Equilavent to ``__name__ == '__main__'``."
    return inspect.stack()[frame][0].f_globals['__name__'] == '__main__'

def _call_if_main(frame, f, args):
    if is_main(frame): return f(*args)

def call_if_main(f,*args):
    "Call the `f` with `args` if the caller's module is main."
    return _call_if_main(3,f,args)

def run_main(f,*args):
    "Call `f` with the `args` and terminate the program with its return code if the caller's module is main."
    sys.exit(_call_if_main(3,f,args))

def compare_and_swap(var, compare, new):
    "If `var` is equal to `compare`, set it to `new`."
    if assign('v', inspect.stack()[1][0].f_globals)[var] == compare:
        v[var] = new
